#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Example property package for the saponification of Ethyl Acetate with NaOH
Assumes dilute solutions with properties of H2O.
"""
# TODO: Missing docstrings
# pylint: disable=missing-function-docstring

# Import Pyomo libraries
from pyomo.environ import Constraint, exp, Param, Set, units, Var

# Import IDAES cores
from idaes.core import (
    declare_process_block_class,
    MaterialFlowBasis,
    ReactionParameterBlock,
    ReactionBlockDataBase,
    ReactionBlockBase,
)
from idaes.core.util.misc import add_object_reference
from idaes.core.util.constants import Constants as const
import idaes.logger as idaeslog
from idaes.core.scaling import CustomScalerBase


# Some more information about this module
__author__ = "Andrew Lee"

from idaes.core.util.scaling import get_scaling_factor

# Set up logger
_log = idaeslog.getLogger(__name__)


@declare_process_block_class("SaponificationReactionParameterBlock")
class ReactionParameterData(ReactionParameterBlock):
    """
    Property Parameter Block Class

    Contains parameters and indexing sets associated with properties for
    superheated steam.

    """

    def build(self):
        """
        Callable method for Block construction.
        """
        super(ReactionParameterData, self).build()

        self._reaction_block_class = ReactionBlock

        # Reaction Index
        self.rate_reaction_idx = Set(initialize=["R1"])

        # Reaction Stoichiometry
        self.rate_reaction_stoichiometry = {
            ("R1", "Liq", "NaOH"): -1,
            ("R1", "Liq", "EthylAcetate"): -1,
            ("R1", "Liq", "SodiumAcetate"): 1,
            ("R1", "Liq", "Ethanol"): 1,
            ("R1", "Liq", "H2O"): 0,
        }

        # Arrhenius Constant
        self.arrhenius = Param(
            default=3.132e6,
            doc="Arrhenius constant",
            units=units.m**3 / units.mol / units.s,
        )

        # Activation Energy
        self.energy_activation = Param(
            default=43000, doc="Activation energy [J/mol]", units=units.J / units.mol
        )

        # Heat of Reaction
        dh_rxn_dict = {"R1": -49000}
        self.dh_rxn = Param(
            self.rate_reaction_idx,
            initialize=dh_rxn_dict,
            doc="Heat of reaction [J/mol]",
            units=units.J / units.mol,
        )

    @classmethod
    def define_metadata(cls, obj):
        obj.add_properties(
            {
                "k_rxn": {"method": "_rate_constant", "units": "m^3/mol.s"},
                "reaction_rate": {"method": "_rxn_rate", "units": "mol/m^3.s"},
            }
        )
        obj.add_default_units(
            {
                "time": units.s,
                "length": units.m,
                "mass": units.kg,
                "amount": units.mol,
                "temperature": units.K,
            }
        )


class SaponificationReactionScaler(CustomScalerBase):
    """
    Scaler for saponification reaction package.

    Variables are scaled by nominal order of magnitude, and constraints
    using the inverse maximum scheme.
    """

    DEFAULT_SCALING_FACTORS = {"reaction_rate": 1e2}

    def variable_scaling_routine(
        self, model, overwrite: bool = False, submodel_scalers: dict = None
    ):
        if model.is_property_constructed("k_rxn"):
            # First check to see if k_rxn is already scaled
            sf = get_scaling_factor(model.k_rxn)

            if sf is not None and not overwrite:
                # k_rxn already has a scaling factor and we are not set to overwrite - move on
                pass
            else:
                # Hopefully temperature has been scaled, so we can get the nominal value of k_rxn
                # by walking the expression in the constraint.
                nominals = self.get_expression_nominal_values(model.arrhenius_eqn)

                # We should get two values, k_rxn (LHS) and the Arrhenius equation (RHS)
                # As of 10/3/2024, the LHS will be the 0-th element of the list, and the RHS the 1st
                # However, we cannot assume this will always be the case

                # If LHS has been scaled, nominal will be 1/sf, otherwise it will be k_rxn.value
                # Find the value which does NOT match this - guess that this is the 1st element
                if nominals[1] != model.k_rxn.value and sf is None:
                    # This is the most likely case, so check it first
                    nominal = nominals[1]
                elif sf is not None and nominals[1] != 1 / sf:
                    # Next, check for case where k_rxn was already scaled
                    nominal = nominals[1]
                else:
                    # Otherwise we have the case where something changed in Pyomo since 10/3/2024
                    nominal = nominals[0]

                self.set_variable_scaling_factor(
                    model.k_rxn,
                    1 / nominal,
                    overwrite=overwrite,
                )

        if model.is_property_constructed("reaction_rate"):
            for j in model.reaction_rate.values():
                self.scale_variable_by_default(j, overwrite=overwrite)

    def constraint_scaling_routine(
        self, model, overwrite: bool = False, submodel_scalers: dict = None
    ):
        if model.is_property_constructed("arrhenius_eqn"):
            self.scale_constraint_by_nominal_value(
                model.arrhenius_eqn,
                scheme="inverse_maximum",
                overwrite=overwrite,
            )

        if model.is_property_constructed("rate_expression"):
            for j in model.rate_expression.values():
                self.scale_constraint_by_nominal_value(
                    j,
                    scheme="inverse_maximum",
                    overwrite=overwrite,
                )


class _ReactionBlock(ReactionBlockBase):
    """
    This Class contains methods which should be applied to Reaction Blocks as a
    whole, rather than individual elements of indexed Reaction Blocks.
    """

    default_scaler = SaponificationReactionScaler

    def initialize(blk, outlvl=idaeslog.NOTSET, **kwargs):
        """
        Initialization routine for reaction package.

        Keyword Arguments:
            outlvl : sets output level of initialization routine

        Returns:
            None
        """
        init_log = idaeslog.getInitLogger(blk.name, outlvl, tag="properties")
        init_log.info("Initialization Complete.")


@declare_process_block_class("ReactionBlock", block_class=_ReactionBlock)
class ReactionBlockData(ReactionBlockDataBase):
    """
    An example reaction package for saponification of ethyl acetate
    """

    def build(self):
        """
        Callable method for Block construction
        """
        super(ReactionBlockData, self).build()

        # Create references to state vars
        # Concentration
        add_object_reference(self, "conc_mol_comp_ref", self.state_ref.conc_mol_comp)

        # Temperature
        add_object_reference(self, "temperature_ref", self.state_ref.temperature)

        # Heat of reaction - no _ref as this is the actual property
        add_object_reference(self, "dh_rxn", self.config.parameters.dh_rxn)

    # Rate constant method
    def _rate_constant(self):
        self.k_rxn = Var(
            initialize=1,
            doc="Rate constant [m^3/mol.s]",
            units=units.m**3 / units.mol / units.s,
        )

        try:
            self.arrhenius_eqn = Constraint(
                expr=self.k_rxn
                == self.params.arrhenius
                * exp(
                    -self.params.energy_activation
                    / (const.gas_constant * self.temperature_ref)
                )
            )
        except AttributeError:
            # If constraint fails, clean up so that DAE can try again later
            self.del_component(self.k_rxn)
            self.del_component(self.arrhenius_eqn)
            raise

    # Rate of reaction method
    def _rxn_rate(self):
        self.reaction_rate = Var(
            self.params.rate_reaction_idx,
            initialize=0,
            doc="Rate of reaction [mol/m^3.s]",
            units=units.mol / units.m**3 / units.s,
        )

        try:

            def rate_rule(b, r):
                return b.reaction_rate[r] == (
                    b.k_rxn
                    * b.conc_mol_comp_ref["EthylAcetate"]
                    * b.conc_mol_comp_ref["NaOH"]
                )

            self.rate_expression = Constraint(
                self.params.rate_reaction_idx, rule=rate_rule
            )
        except AttributeError:
            # If constraint fails, clean up so that DAE can try again later
            self.del_component(self.reaction_rate)
            self.del_component(self.rate_expression)
            raise

    def get_reaction_rate_basis(b):
        return MaterialFlowBasis.molar

    def model_check(blk):
        pass



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Downcomer model
Main assumptions:
* Heat is given (zero if adiabatic)
* Calculate pressure change due to friction and gravity
* Assume enthalpy_in == enthalpy_out + heat

Created on Aug 27, 2020 by Boiler Team (J. Ma, M. Zamarripa)
"""
import pytest

# Import Pyomo libraries
import pyomo.environ as pyo
from pyomo.util.check_units import assert_units_consistent

# Import IDAES core
from idaes.core import FlowsheetBlock
from idaes.core.util.model_statistics import degrees_of_freedom

# Import Unit Model Modules
from idaes.models.properties import iapws95

from idaes.core.util.testing import initialization_tester
from idaes.core.solvers import get_solver
from idaes.models_extra.power_generation.unit_models.downcomer import Downcomer

# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_solver()

# -----------------------------------------------------------------------------


@pytest.fixture(scope="module")
def build_downcomer():
    m = pyo.ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.properties = iapws95.Iapws95ParameterBlock()
    m.fs.unit = Downcomer(
        property_package=m.fs.properties, has_holdup=False, has_heat_transfer=True
    )
    return m


@pytest.mark.skipif(not iapws95.iapws95_available(), reason="IAPWS not available")
@pytest.mark.unit
def test_basic_build(build_downcomer):
    """Make a model and make sure it doesn't throw exception"""
    m = build_downcomer
    assert degrees_of_freedom(m) == 7
    # Check unit config arguments
    assert len(m.fs.unit.config) == 8
    assert m.fs.unit.config.has_heat_transfer
    assert isinstance(m.fs.unit.heat_duty, pyo.Var)
    assert isinstance(m.fs.unit.deltaP, pyo.Var)
    assert m.fs.unit.config.property_package is m.fs.properties


@pytest.mark.skipif(not iapws95.iapws95_available(), reason="IAPWS not available")
@pytest.mark.integration
def test_units(build_downcomer):
    assert_units_consistent(build_downcomer)


@pytest.mark.skipif(not iapws95.iapws95_available(), reason="IAPWS not available")
@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.component
def test_initialize_unit(build_downcomer):
    m = build_downcomer
    m.fs.unit.diameter.fix(0.3)
    m.fs.unit.number_downcomers.fix(6)  # number of downcomers
    m.fs.unit.height.fix(40.1)
    m.fs.unit.heat_duty[:].fix(0.0)

    # NETL Baseline values
    m.fs.unit.inlet.enth_mol.fix(24944.7)
    m.fs.unit.inlet.flow_mol.fix(49552.8)
    m.fs.unit.inlet.pressure.fix(12025072.9)

    state_args = {"flow_mol": 49552.8, "pressure": 12025072.9, "enth_mol": 24944.7}
    initialization_tester(build_downcomer, dof=0, state_args=state_args)


@pytest.mark.skipif(not iapws95.iapws95_available(), reason="IAPWS not available")
@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.component
def test_solve_unit(build_downcomer):
    m = build_downcomer

    m.fs.unit.inlet.enth_mol.fix()
    m.fs.unit.inlet.flow_mol.fix()
    m.fs.unit.inlet.pressure.fix()
    assert degrees_of_freedom(m) == 0

    results = solver.solve(m, tee=True)
    # Check for optimal solution
    assert pyo.check_optimal_termination(results)

    # check material balance
    assert (
        pytest.approx(
            pyo.value(
                m.fs.unit.control_volume.properties_in[0].flow_mol
                - m.fs.unit.control_volume.properties_out[0].flow_mol
            ),
            abs=1e-3,
        )
        == 0
    )

    # pressure drop
    assert pytest.approx(273583.120306, rel=1e-5) == pyo.value(m.fs.unit.deltaP[0])
    # check energy balance
    assert pytest.approx(
        pyo.value(m.fs.unit.control_volume.properties_in[0].enth_mol), abs=1e-3
    ) == pyo.value(m.fs.unit.control_volume.properties_out[0].enth_mol)

    assert pytest.approx(580.83299, abs=1e-3) == pyo.value(
        m.fs.unit.control_volume.properties_out[0].temperature
    )



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
import logging
import os


def level_num(name, default_level):
    """Get level num from name"""
    level_mapping = dict(
        ERROR=logging.ERROR,
        INFO=logging.INFO,
        WARNING=logging.WARNING,
        DEBUG=logging.DEBUG,
    )
    return level_mapping.get(name.upper(), default_level)


# set up test logger
_log = logging.getLogger(__name__)

level = logging.WARNING
if "IDAES_TEST_LOG_LEVEL" in os.environ:
    env_level = os.environ["IDAES_TEST_LOG_LEVEL"]
    level = level_num(env_level, level)

_h = logging.StreamHandler()
_h.setFormatter(
    logging.Formatter(
        "%(asctime)s [%(levelname)s] " "%(filename)s:%(lineno)d :: %(message)s"
    )
)
_log.addHandler(_h)
_log.setLevel(level)
_log.propagate = False



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
This module contains heat utility minimization functions for use in
IDAES models.
"""

__author__ = "Alejandro Garciadiego"

import pytest
from pyomo.environ import (
    check_optimal_termination,
    ConcreteModel,
    Objective,
    units as pyunits,
)
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.solvers import get_solver

from idaes.core import VaporPhase, Component
from idaes.models.properties.modular_properties.state_definitions import FTPx
from idaes.models.properties.modular_properties.eos.ideal import Ideal
import idaes.models.properties.modular_properties.pure.RPP4 as RPP4

from idaes.models.unit_models import Heater
from idaes.core import FlowsheetBlock

from idaes.models.properties.modular_properties.base.generic_property import (
    GenericParameterBlock,
)

from idaes.core.util.utility_minimization import (
    min_utility,
    PinchDataClass,
    heat_ex_data,
    gen_curves,
)


# -----------------------------------------------------------------------------
# [1] systematic Methods of Chemical Process Design (1997)
#     Chemical Engineering Series - L. T. Biegler, I. E. Grossmann,
#     A. W. Westerberg, page 529, Example 16.1
# Get default solver for testing
solver = get_solver("ipopt_v2")


@pytest.mark.unit
def test_PinchDataClass():

    PD = PinchDataClass(100, 80)
    PD.initQAh = {"fs.H101": 540756, "fs.H102": 719998, "fs.H103": 0, "fs.H104": 59787}
    PD.initQAc = {
        "fs.H101": 477515,
        "fs.H102": 554999,
        "fs.H103": 29875,
        "fs.H104": 119502,
    }

    assert PD.initQs == 100
    assert PD.initQw == 80
    assert PD.initQAh == {
        "fs.H101": 540756,
        "fs.H102": 719998,
        "fs.H103": 0,
        "fs.H104": 59787,
    }
    assert PD.initQAc == {
        "fs.H101": 477515,
        "fs.H102": 554999,
        "fs.H103": 29875,
        "fs.H104": 119502,
    }


# Author: Alejandro Garciadiego
class TestStateBlock(object):
    @pytest.fixture(scope="class")
    def model(self):
        configuration = {
            # Specifying components
            "components": {
                "A": {
                    "type": Component,
                    "enth_mol_ig_comp": RPP4,
                    "entr_mol_ig_comp": RPP4,
                    "parameter_data": {
                        "mw": (2.016e-3, pyunits.kg / pyunits.mol),  # [1]
                        "pressure_crit": (12.9e5, pyunits.Pa),  # [1]
                        "temperature_crit": (33.2, pyunits.K),  # [1]
                        "omega": -0.218,
                        "cp_mol_ig_comp_coeff": {
                            "A": (1000000, pyunits.J / pyunits.mol / pyunits.K),
                            "B": (0, pyunits.J / pyunits.mol / pyunits.K**2),
                            "C": (0, pyunits.J / pyunits.mol / pyunits.K**3),
                            "D": (0, pyunits.J / pyunits.mol / pyunits.K**4),
                        },
                        "entr_mol_form_vap_comp_ref": (
                            0,
                            pyunits.J / pyunits.mol / pyunits.K,
                        ),
                        "enth_mol_form_vap_comp_ref": (0.0, pyunits.J / pyunits.mol),
                    },
                }
            },
            # Specifying phases
            "phases": {"Vap": {"type": VaporPhase, "equation_of_state": Ideal}},
            # Set base units of measurement
            "base_units": {
                "time": pyunits.s,
                "length": pyunits.m,
                "mass": pyunits.kg,
                "amount": pyunits.mol,
                "temperature": pyunits.K,
            },
            # Specifying state definition
            "state_definition": FTPx,
            "state_bounds": {
                "flow_mol": (0, 100, 1000, pyunits.mol / pyunits.s),
                "temperature": (100, 300, 450, pyunits.K),
                "pressure": (5e4, 1e5, 1e6, pyunits.Pa),
            },
            "pressure_ref": (1e5, pyunits.Pa),
            "temperature_ref": (300, pyunits.K),
        }

        # Create the ConcreteModel and the FlowsheetBlock, and attach the
        # flowsheet block to it.
        model = ConcreteModel()

        model.fs = FlowsheetBlock(dynamic=False)

        # Add properties parameter blocks to the flowsheet with specifications
        model.fs.props = GenericParameterBlock(**configuration)

        # Create an instance of the units, attaching them to the flowsheet
        # Specify that the property package to be used with with eash unit.
        model.fs.H101 = Heater(
            property_package=model.fs.props,
            has_pressure_change=False,
            has_phase_equilibrium=False,
        )

        model.fs.H102 = Heater(
            property_package=model.fs.props,
            has_pressure_change=False,
            has_phase_equilibrium=False,
        )

        model.fs.H103 = Heater(
            property_package=model.fs.props,
            has_pressure_change=False,
            has_phase_equilibrium=False,
        )

        model.fs.H104 = Heater(
            property_package=model.fs.props,
            has_pressure_change=False,
            has_phase_equilibrium=False,
        )

        model.fs.H101.inlet.mole_frac_comp[0, "A"].fix(1)
        model.fs.H101.inlet.flow_mol[0].fix(1.5)
        model.fs.H101.inlet.temperature.fix(160)
        model.fs.H101.inlet.pressure.fix(100000)

        model.fs.H101.outlet.temperature.fix(400)

        model.fs.H102.inlet.mole_frac_comp[0, "A"].fix(1)
        model.fs.H102.inlet.flow_mol[0].fix(1.3)
        model.fs.H102.inlet.temperature.fix(100)
        model.fs.H102.inlet.pressure.fix(100000)

        model.fs.H102.outlet.temperature.fix(250)

        model.fs.H103.inlet.mole_frac_comp[0, "A"].fix(1)
        model.fs.H103.inlet.flow_mol[0].fix(1)
        model.fs.H103.inlet.temperature.fix(400)
        model.fs.H103.inlet.pressure.fix(100000)

        model.fs.H103.outlet.temperature.fix(120)

        model.fs.H104.inlet.mole_frac_comp[0, "A"].fix(1)
        model.fs.H104.inlet.flow_mol[0].fix(2)
        model.fs.H104.inlet.temperature.fix(340)
        model.fs.H104.inlet.pressure.fix(100000)

        model.fs.H104.outlet.temperature.fix(120)

        assert degrees_of_freedom(model) == 0

        return model

    @pytest.mark.unit
    def test_dof(self, model):

        # add heat integration constraints
        min_utility(
            model.fs, [model.fs.H101, model.fs.H102], [model.fs.H103, model.fs.H104], 20
        )

        assert degrees_of_freedom(model) == 1

        model.fs.objective = Objective(expr=model.fs.Qs + model.fs.Qw)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.unit
    def test_solve(self, model):
        results = solver.solve(model)

        # Check for optimal solution
        assert check_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.unit
    def test_solution(self, model):
        assert model.fs.QAh["fs.H101"].value == pytest.approx(540, abs=1e-0)
        assert model.fs.QAh["fs.H102"].value == pytest.approx(719, abs=1e-0)
        assert model.fs.QAh["fs.H103"].value == pytest.approx(0, abs=1e-0)
        assert model.fs.QAh["fs.H104"].value == pytest.approx(60, abs=1e-0)
        assert model.fs.QAc["fs.H101"].value == pytest.approx(476, abs=1e-0)
        assert model.fs.QAc["fs.H102"].value == pytest.approx(554, abs=1e-0)
        assert model.fs.QAc["fs.H103"].value == pytest.approx(30, abs=1e-0)
        assert model.fs.QAc["fs.H104"].value == pytest.approx(120, abs=1e-0)
        assert model.fs.Qs.value == pytest.approx(60, abs=1e-0)
        assert model.fs.Qw.value == pytest.approx(225, abs=1e-0)

    @pytest.mark.unit
    def test_heat_ex_data(self, model):

        CD = heat_ex_data(
            model.fs, [model.fs.H101, model.fs.H102], [model.fs.H103, model.fs.H104]
        )

        assert CD.Cooling_Tin[0] == 401
        assert CD.Cooling_Tin[1] == 341
        assert CD.Cooling_Tout[0] == 120
        assert CD.Cooling_Tout[1] == 120
        assert CD.Cooling_Q[0] == -280
        assert CD.Cooling_Q[1] == -440
        assert CD.Heating_Tin[0] == 160
        assert CD.Heating_Tin[1] == 100
        assert CD.Heating_Tout[0] == 401
        assert CD.Heating_Tout[1] == 251
        assert CD.Heating_Q[0] == 360
        assert CD.Heating_Q[1] == 195

    @pytest.mark.unit
    def test_curve_data(self):

        T_test, Q_test = gen_curves([401, 341], [120, 120], [-280, -440])
        Ttest = [120, 120, 341, 401]
        for i in range(len(Ttest)):
            assert T_test[i] == Ttest[i]
        Qtest = [0, 0, -660.213, -720]
        for i in range(len(Qtest)):
            assert Q_test[i] == pytest.approx(Qtest[i], abs=1e-3)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests for methods from CoolProp wrapper

Authors: Andrew Lee
"""

import pytest
from numpy import arange

from pyomo.environ import (
    Block,
    ConcreteModel,
    Param,
    SolverStatus,
    TerminationCondition,
    units as pyunits,
    value,
    Var,
)
from pyomo.util.check_units import assert_units_equivalent

from idaes.core import FlowsheetBlock, Component, LiquidPhase, VaporPhase
from idaes.models.properties.modular_properties.base.generic_property import (
    GenericParameterBlock,
)
from idaes.models.properties.modular_properties.state_definitions import FTPx
from idaes.models.properties.modular_properties.eos.ceos import Cubic, CubicType
from idaes.core.solvers import get_solver
from idaes.models.properties.modular_properties.pure.ConstantProperties import Constant


CoolProp = pytest.importorskip("CoolProp.CoolProp", reason="CoolProp not installed")

from idaes.models.properties.modular_properties.coolprop.coolprop_wrapper import (
    CoolPropWrapper,
    CoolPropExpressionError,
    CoolPropPropertyError,
)

solver = get_solver("ipopt_v2")


class TestWrapper:
    @pytest.mark.unit
    def test_load_component(self):
        # Clear cached components to be sure
        CoolPropWrapper._cached_components = {}

        # Load parameters for oxygen
        prop_dict = CoolPropWrapper._load_component_data("Oxygen")

        assert prop_dict is not None
        assert "Oxygen" in CoolPropWrapper._cached_components
        assert CoolPropWrapper._cached_components["Oxygen"] is prop_dict
        assert isinstance(prop_dict, dict)
        for k in ["STATES", "ANCILLARIES"]:
            assert k in prop_dict

    @pytest.mark.unit
    def test_load_component_invalid(self):
        with pytest.raises(
            RuntimeError,
            match="Failed to find component foo in CoolProp " "JSON database.",
        ):
            CoolPropWrapper._load_component_data("foo")

    @pytest.mark.unit
    def test_get_component(self):
        prop_dict = CoolPropWrapper._get_component_data("Oxygen")

        assert CoolPropWrapper._cached_components["Oxygen"] is prop_dict

    @pytest.mark.unit
    def test_get_component_alias(self):
        # Load oxygen properties using aliases
        prop_dict = CoolPropWrapper._get_component_data("R732")

        assert CoolPropWrapper._cached_components["Oxygen"] is prop_dict
        assert "R732" in CoolPropWrapper._cached_components
        assert (
            CoolPropWrapper._cached_components["R732"]
            is CoolPropWrapper._cached_components["Oxygen"]
        )

    @pytest.mark.unit
    def test_load_component_by_alias(self):
        # Load CO2 data using one of its aliases
        prop_dict = CoolPropWrapper._load_component_data("R744")

        assert CoolPropWrapper._cached_components["R744"] is prop_dict
        assert "R744" in CoolPropWrapper._cached_components

        # Retrieve CO2 data using its normal name
        prop_dict2 = CoolPropWrapper._get_component_data("CO2")

        assert prop_dict2 is prop_dict
        assert "CO2" in CoolPropWrapper._cached_components
        assert (
            CoolPropWrapper._cached_components["CO2"]
            is CoolPropWrapper._cached_components["R744"]
        )

    @pytest.mark.unit
    def test_get_component_invalid(self):
        with pytest.raises(
            RuntimeError,
            match="Failed to find component foo in CoolProp " "JSON database.",
        ):
            CoolPropWrapper._get_component_data("foo")

    @pytest.mark.unit
    def test_flush_cached_components(self):
        CoolPropWrapper.flush_cached_components()

        assert CoolPropWrapper._cached_components == {}

    @pytest.mark.unit
    def test_get_critical_properties(self):
        Tc = CoolPropWrapper._get_critical_property("Acetone", "T")
        assert Tc == (508.1, pyunits.K)

        Pc = CoolPropWrapper._get_critical_property("Acetone", "p")
        assert Pc == (4700000.0, pyunits.Pa)

        rhoc = CoolPropWrapper._get_critical_property("Acetone", "rhomolar")
        assert rhoc[0] == 4699.999999999999
        assert_units_equivalent(rhoc[1], pyunits.mol / pyunits.m**3)

        hc = CoolPropWrapper._get_critical_property("Acetone", "hmolar")
        assert hc[0] == 31614.73051047263
        assert_units_equivalent(hc[1], pyunits.J / pyunits.mol)

        sc = CoolPropWrapper._get_critical_property("Acetone", "smolar")
        assert sc[0] == 72.97112978635582
        assert_units_equivalent(sc[1], pyunits.J / pyunits.mol / pyunits.K)

    @pytest.mark.unit
    def test_get_eos_properties(self):
        omega = CoolPropWrapper._get_eos_property("Acetone", "acentric")
        assert omega == (0.3071, pyunits.dimensionless)

        mw = CoolPropWrapper._get_eos_property("Acetone", "molar_mass")
        assert mw[0] == 0.05807914
        assert_units_equivalent(mw[1], pyunits.kg / pyunits.mol)

    @pytest.mark.unit
    def test_get_parameter_value(self):
        Tc = CoolPropWrapper.get_parameter_value("Acetone", "temperature_crit")
        assert Tc == (508.1, pyunits.K)

        Pc = CoolPropWrapper.get_parameter_value("Acetone", "pressure_crit")
        assert Pc == (4700000.0, pyunits.Pa)

        rhoc = CoolPropWrapper.get_parameter_value("Acetone", "dens_mol_crit")
        assert rhoc[0] == 4699.999999999999
        assert_units_equivalent(rhoc[1], pyunits.mol / pyunits.m**3)

        hc = CoolPropWrapper.get_parameter_value("Acetone", "enth_mol_crit")
        assert hc[0] == 31614.73051047263
        assert_units_equivalent(hc[1], pyunits.J / pyunits.mol)

        sc = CoolPropWrapper.get_parameter_value("Acetone", "entr_mol_crit")
        assert sc[0] == 72.97112978635582
        assert_units_equivalent(sc[1], pyunits.J / pyunits.mol / pyunits.K)

        omega = CoolPropWrapper.get_parameter_value("Acetone", "omega")
        assert omega == (0.3071, pyunits.dimensionless)

        mw = CoolPropWrapper.get_parameter_value("Acetone", "mw")
        assert mw[0] == 0.05807914
        assert_units_equivalent(mw[1], pyunits.kg / pyunits.mol)

    @pytest.mark.unit
    def test_pressure_sat_no_params(self):
        m = ConcreteModel()
        # Dummy a Component object with the name of a property which does not
        # have data for saturation pressure
        m.Air = Block()

        with pytest.raises(
            CoolPropPropertyError,
            match="Could not retrieve parameters for "
            "pressure_sat of component Air from CoolProp. This "
            "likely indicates that CoolProp does not have "
            "values for the necessary parameters.",
        ):
            CoolPropWrapper.pressure_sat_comp.build_parameters(m.Air)

    @pytest.mark.unit
    def test_pressure_sat_unrecognised_form(self):
        m = ConcreteModel()
        # First, add a dummy component to the cached components which uses
        # unsupoorted form
        CoolPropWrapper._cached_components["TestComp"] = {
            "ANCILLARIES": {"pS": {"type": "foo", "using_tau_r": True}},
            "INFO": {"ALIASES": ["testcomp"], "NAME": "TestComp"},
        }
        m.TestComp = Block()

        with pytest.raises(
            CoolPropExpressionError,
            match="Found unsupported expression form for "
            "pressure_sat of component TestComp. This likely "
            "occurred due to changes in CoolProp and the "
            "interface should be updated.",
        ):
            CoolPropWrapper.pressure_sat_comp.build_parameters(m.TestComp)

        # Pressure_sat uses has two parts to form. Set type to supported form
        # and using_tau_r to False (unsupported)
        CoolPropWrapper._cached_components["TestComp"]["ANCILLARIES"]["pS"][
            "type"
        ] = "pL"
        CoolPropWrapper._cached_components["TestComp"]["ANCILLARIES"]["pS"][
            "using_tau_r"
        ] = False
        with pytest.raises(
            CoolPropExpressionError,
            match="Found unsupported expression form for "
            "pressure_sat of component TestComp. This likely "
            "occurred due to changes in CoolProp and the "
            "interface should be updated.",
        ):
            CoolPropWrapper.pressure_sat_comp.build_parameters(m.TestComp)


class TestCoolPropProperties(object):
    @pytest.fixture(scope="class")
    def m(self):
        # Clear cached components to ensure clean slate
        CoolPropWrapper.flush_cached_components()

        m = ConcreteModel()

        m.fs = FlowsheetBlock(dynamic=False)

        configuration = {
            # Specifying components
            "components": {
                "benzene": {
                    "type": Component,
                    "dens_mol_liq_comp": CoolPropWrapper,
                    "enth_mol_liq_comp": CoolPropWrapper,
                    "enth_mol_ig_comp": CoolPropWrapper,
                    "entr_mol_liq_comp": CoolPropWrapper,
                    "entr_mol_ig_comp": CoolPropWrapper,
                    "pressure_sat_comp": CoolPropWrapper,
                    "parameter_data": {
                        "mw": CoolPropWrapper,
                        "dens_mol_crit": CoolPropWrapper,
                        "pressure_crit": CoolPropWrapper,
                        "temperature_crit": CoolPropWrapper,
                        "omega": CoolPropWrapper,
                    },
                }
            },
            # Specifying phases
            "phases": {
                "Liq": {
                    "type": LiquidPhase,
                    "equation_of_state": Cubic,
                    "equation_of_state_options": {"type": CubicType.PR},
                }
            },
            # Set base units of measurement
            "base_units": {
                "time": pyunits.s,
                "length": pyunits.m,
                "mass": pyunits.kg,
                "amount": pyunits.mol,
                "temperature": pyunits.K,
            },
            # Specifying state definition
            "state_definition": FTPx,
            "state_bounds": {
                "flow_mol": (0, 100, 1000, pyunits.mol / pyunits.s),
                "temperature": (273.15, 300, 500, pyunits.K),
                "pressure": (5e4, 1e5, 1e6, pyunits.Pa),
            },
            "pressure_ref": (101325, pyunits.Pa),
            "temperature_ref": (298.15, pyunits.K),
            "parameter_data": {"PR_kappa": {("benzene", "benzene"): 0.000}},
        }

        m.fs.props = GenericParameterBlock(**configuration)

        m.fs.state = m.fs.props.build_state_block([0], defined_state=True)

        m.fs.state[0].flow_mol.fix(1)
        m.fs.state[0].pressure.fix(101325)
        m.fs.state[0].temperature.fix(300)
        m.fs.state[0].mole_frac_comp["benzene"].fix(1)

        return m

    @pytest.mark.unit
    def test_physical_constants(self, m):
        # Benzene parameters
        assert isinstance(m.fs.props.benzene.temperature_crit, Var)
        assert m.fs.props.benzene.temperature_crit.fixed
        assert value(m.fs.props.benzene.temperature_crit) == CoolProp.PropsSI(
            "TCRIT", "Benzene"
        )

        assert isinstance(m.fs.props.benzene.pressure_crit, Var)
        assert m.fs.props.benzene.pressure_crit.fixed
        assert value(m.fs.props.benzene.pressure_crit) == CoolProp.PropsSI(
            "PCRIT", "Benzene"
        )

        assert isinstance(m.fs.props.benzene.dens_mol_crit, Var)
        assert m.fs.props.benzene.dens_mol_crit.fixed
        assert value(m.fs.props.benzene.dens_mol_crit) == CoolProp.PropsSI(
            "RHOMOLAR_CRITICAL", "Benzene"
        )

        assert isinstance(m.fs.props.benzene.mw, Param)
        assert value(m.fs.props.benzene.mw) == CoolProp.PropsSI("molarmass", "Benzene")

        assert isinstance(m.fs.props.benzene.omega, Var)
        assert m.fs.props.benzene.omega.fixed
        assert value(m.fs.props.benzene.omega) == CoolProp.PropsSI(
            "acentric", "Benzene"
        )

    @pytest.mark.unit
    def test_dens_mol_liq_comp(self, m):
        assert isinstance(m.fs.props.benzene.dens_mol_liq_comp_coeff_n1, Var)
        assert isinstance(m.fs.props.benzene.dens_mol_liq_comp_coeff_n2, Var)
        assert isinstance(m.fs.props.benzene.dens_mol_liq_comp_coeff_n3, Var)
        assert isinstance(m.fs.props.benzene.dens_mol_liq_comp_coeff_n4, Var)
        assert isinstance(m.fs.props.benzene.dens_mol_liq_comp_coeff_n5, Var)
        assert isinstance(m.fs.props.benzene.dens_mol_liq_comp_coeff_n6, Var)

        assert isinstance(m.fs.props.benzene.dens_mol_liq_comp_coeff_t1, Var)
        assert isinstance(m.fs.props.benzene.dens_mol_liq_comp_coeff_t2, Var)
        assert isinstance(m.fs.props.benzene.dens_mol_liq_comp_coeff_t3, Var)
        assert isinstance(m.fs.props.benzene.dens_mol_liq_comp_coeff_t4, Var)
        assert isinstance(m.fs.props.benzene.dens_mol_liq_comp_coeff_t5, Var)
        assert isinstance(m.fs.props.benzene.dens_mol_liq_comp_coeff_t6, Var)

        assert m.fs.props.benzene.dens_mol_liq_comp_coeff_n1.value == pytest.approx(
            2.852587673922022, rel=1e-10
        )
        assert m.fs.props.benzene.dens_mol_liq_comp_coeff_n1.fixed
        assert m.fs.props.benzene.dens_mol_liq_comp_coeff_n2.value == pytest.approx(
            -0.5596547795188646, rel=1e-10
        )
        assert m.fs.props.benzene.dens_mol_liq_comp_coeff_n2.fixed
        assert m.fs.props.benzene.dens_mol_liq_comp_coeff_n3.value == pytest.approx(
            14.872052666571532, rel=1e-10
        )
        assert m.fs.props.benzene.dens_mol_liq_comp_coeff_n3.fixed
        assert m.fs.props.benzene.dens_mol_liq_comp_coeff_n4.value == pytest.approx(
            -66.42959110979461, rel=1e-10
        )
        assert m.fs.props.benzene.dens_mol_liq_comp_coeff_n4.fixed
        assert m.fs.props.benzene.dens_mol_liq_comp_coeff_n5.value == pytest.approx(
            1158.1329856052375, rel=1e-10
        )
        assert m.fs.props.benzene.dens_mol_liq_comp_coeff_n5.fixed
        assert m.fs.props.benzene.dens_mol_liq_comp_coeff_n6.value == pytest.approx(
            -3128.774352224071, rel=1e-10
        )
        assert m.fs.props.benzene.dens_mol_liq_comp_coeff_n6.fixed

        assert m.fs.props.benzene.dens_mol_liq_comp_coeff_t1.value == pytest.approx(
            0.407, rel=1e-10
        )
        assert m.fs.props.benzene.dens_mol_liq_comp_coeff_t1.fixed
        assert m.fs.props.benzene.dens_mol_liq_comp_coeff_t2.value == pytest.approx(
            0.565, rel=1e-10
        )
        assert m.fs.props.benzene.dens_mol_liq_comp_coeff_t2.fixed
        assert m.fs.props.benzene.dens_mol_liq_comp_coeff_t3.value == pytest.approx(
            4.029, rel=1e-10
        )
        assert m.fs.props.benzene.dens_mol_liq_comp_coeff_t3.fixed
        assert m.fs.props.benzene.dens_mol_liq_comp_coeff_t4.value == pytest.approx(
            5.699, rel=1e-10
        )
        assert m.fs.props.benzene.dens_mol_liq_comp_coeff_t4.fixed
        assert m.fs.props.benzene.dens_mol_liq_comp_coeff_t5.value == pytest.approx(
            9.989, rel=1e-10
        )
        assert m.fs.props.benzene.dens_mol_liq_comp_coeff_t5.fixed
        assert m.fs.props.benzene.dens_mol_liq_comp_coeff_t6.value == pytest.approx(
            12.299, rel=1e-10
        )
        assert m.fs.props.benzene.dens_mol_liq_comp_coeff_t6.fixed

        # CoolProp results are non-ideal, which results in deviations
        for T in range(300, 401, 10):
            m.fs.state[0].temperature.fix(T)
            assert pytest.approx(
                CoolProp.PropsSI("DMOLAR", "T", T, "Q", 0, "benzene"), rel=5e-3
            ) == value(
                CoolPropWrapper.dens_mol_liq_comp.return_expression(
                    m.fs.state[0], m.fs.props.benzene, T * pyunits.K
                )
            )

    @pytest.mark.unit
    def test_enth_mol_liq(self, m):
        assert isinstance(m.fs.props.benzene.enth_mol_liq_comp_coeff_A0, Var)
        assert isinstance(m.fs.props.benzene.enth_mol_liq_comp_coeff_A1, Var)
        assert isinstance(m.fs.props.benzene.enth_mol_liq_comp_coeff_A2, Var)
        assert isinstance(m.fs.props.benzene.enth_mol_liq_comp_coeff_A3, Var)
        assert isinstance(m.fs.props.benzene.enth_mol_liq_comp_coeff_A4, Var)
        assert isinstance(m.fs.props.benzene.enth_mol_liq_comp_coeff_A5, Var)
        assert isinstance(m.fs.props.benzene.enth_mol_liq_comp_coeff_A6, Var)
        assert isinstance(m.fs.props.benzene.enth_mol_liq_comp_coeff_A7, Var)

        assert isinstance(m.fs.props.benzene.enth_mol_liq_comp_coeff_B0, Var)
        assert isinstance(m.fs.props.benzene.enth_mol_liq_comp_coeff_B1, Var)

        assert m.fs.props.benzene.enth_mol_liq_comp_coeff_A0.value == pytest.approx(
            -59920.08975429371, rel=1e-10
        )
        assert m.fs.props.benzene.enth_mol_liq_comp_coeff_A0.fixed
        assert m.fs.props.benzene.enth_mol_liq_comp_coeff_A1.value == pytest.approx(
            -435.3004255527287, rel=1e-10
        )
        assert m.fs.props.benzene.enth_mol_liq_comp_coeff_A1.fixed
        assert m.fs.props.benzene.enth_mol_liq_comp_coeff_A2.value == pytest.approx(
            5.686269274827653, rel=1e-10
        )
        assert m.fs.props.benzene.enth_mol_liq_comp_coeff_A2.fixed
        assert m.fs.props.benzene.enth_mol_liq_comp_coeff_A3.value == pytest.approx(
            -0.026551688405615677, rel=1e-10
        )
        assert m.fs.props.benzene.enth_mol_liq_comp_coeff_A3.fixed
        assert m.fs.props.benzene.enth_mol_liq_comp_coeff_A4.value == pytest.approx(
            7.158706209734695e-05, rel=1e-10
        )
        assert m.fs.props.benzene.enth_mol_liq_comp_coeff_A4.fixed
        assert m.fs.props.benzene.enth_mol_liq_comp_coeff_A5.value == pytest.approx(
            -1.1450245890590754e-07, rel=1e-10
        )
        assert m.fs.props.benzene.enth_mol_liq_comp_coeff_A5.fixed
        assert m.fs.props.benzene.enth_mol_liq_comp_coeff_A6.value == pytest.approx(
            1.0020646255928023e-10, rel=1e-10
        )
        assert m.fs.props.benzene.enth_mol_liq_comp_coeff_A6.fixed
        assert m.fs.props.benzene.enth_mol_liq_comp_coeff_A7.value == pytest.approx(
            -3.714155743547148e-14, rel=1e-10
        )
        assert m.fs.props.benzene.enth_mol_liq_comp_coeff_A7.fixed

        assert m.fs.props.benzene.enth_mol_liq_comp_coeff_B0.value == pytest.approx(
            1, rel=1e-10
        )
        assert m.fs.props.benzene.enth_mol_liq_comp_coeff_B0.fixed
        assert m.fs.props.benzene.enth_mol_liq_comp_coeff_B1.value == pytest.approx(
            -0.0017636924804910123, rel=1e-10
        )
        assert m.fs.props.benzene.enth_mol_liq_comp_coeff_B1.fixed

        assert m.fs.props.benzene.enth_mol_liq_comp_anchor.value == pytest.approx(
            54228.40275249632, rel=1e-10
        )
        assert m.fs.props.benzene.enth_mol_liq_comp_anchor.fixed

        for T in range(300, 401, 10):
            m.fs.state[0].temperature.fix(T)
            assert pytest.approx(
                CoolProp.PropsSI("Hmolar", "T", T, "Q", 0, "benzene"), rel=5e-4
            ) == value(
                CoolPropWrapper.enth_mol_liq_comp.return_expression(
                    m.fs.state[0], m.fs.props.benzene, T * pyunits.K
                )
            )

    @pytest.mark.unit
    def test_enth_mol_ig(self, m):
        assert isinstance(m.fs.props.benzene.enth_mol_ig_comp_coeff_A0, Var)
        assert isinstance(m.fs.props.benzene.enth_mol_ig_comp_coeff_A1, Var)
        assert isinstance(m.fs.props.benzene.enth_mol_ig_comp_coeff_A2, Var)
        assert isinstance(m.fs.props.benzene.enth_mol_ig_comp_coeff_A3, Var)
        assert isinstance(m.fs.props.benzene.enth_mol_ig_comp_coeff_A4, Var)
        assert isinstance(m.fs.props.benzene.enth_mol_ig_comp_coeff_A5, Var)
        assert isinstance(m.fs.props.benzene.enth_mol_ig_comp_coeff_A6, Var)
        assert isinstance(m.fs.props.benzene.enth_mol_ig_comp_coeff_A7, Var)

        assert isinstance(m.fs.props.benzene.enth_mol_ig_comp_coeff_B0, Var)
        assert isinstance(m.fs.props.benzene.enth_mol_ig_comp_coeff_B1, Var)

        assert m.fs.props.benzene.enth_mol_ig_comp_coeff_A0.value == pytest.approx(
            -72325.6503937073, rel=1e-10
        )
        assert m.fs.props.benzene.enth_mol_ig_comp_coeff_A0.fixed
        assert m.fs.props.benzene.enth_mol_ig_comp_coeff_A1.value == pytest.approx(
            2230.356360119971, rel=1e-10
        )
        assert m.fs.props.benzene.enth_mol_ig_comp_coeff_A1.fixed
        assert m.fs.props.benzene.enth_mol_ig_comp_coeff_A2.value == pytest.approx(
            -19.220224456239638, rel=1e-10
        )
        assert m.fs.props.benzene.enth_mol_ig_comp_coeff_A2.fixed
        assert m.fs.props.benzene.enth_mol_ig_comp_coeff_A3.value == pytest.approx(
            0.08579743322690633, rel=1e-10
        )
        assert m.fs.props.benzene.enth_mol_ig_comp_coeff_A3.fixed
        assert m.fs.props.benzene.enth_mol_ig_comp_coeff_A4.value == pytest.approx(
            -0.00022509135257932873, rel=1e-10
        )
        assert m.fs.props.benzene.enth_mol_ig_comp_coeff_A4.fixed
        assert m.fs.props.benzene.enth_mol_ig_comp_coeff_A5.value == pytest.approx(
            3.4964051978046773e-07, rel=1e-10
        )
        assert m.fs.props.benzene.enth_mol_ig_comp_coeff_A5.fixed
        assert m.fs.props.benzene.enth_mol_ig_comp_coeff_A6.value == pytest.approx(
            -2.9861697218437715e-10, rel=1e-10
        )
        assert m.fs.props.benzene.enth_mol_ig_comp_coeff_A6.fixed
        assert m.fs.props.benzene.enth_mol_ig_comp_coeff_A7.value == pytest.approx(
            1.0849858354638855e-13, rel=1e-10
        )
        assert m.fs.props.benzene.enth_mol_ig_comp_coeff_A7.fixed

        assert m.fs.props.benzene.enth_mol_ig_comp_coeff_B0.value == pytest.approx(
            1, rel=1e-10
        )
        assert m.fs.props.benzene.enth_mol_ig_comp_coeff_B0.fixed
        assert m.fs.props.benzene.enth_mol_ig_comp_coeff_B1.value == pytest.approx(
            -0.0017620752364379978, rel=1e-10
        )
        assert m.fs.props.benzene.enth_mol_ig_comp_coeff_B1.fixed

        for T in range(300, 401, 10):
            m.fs.state[0].temperature.fix(T)
            assert pytest.approx(
                CoolProp.PropsSI("Hmolar", "T", T, "Q", 1, "benzene"), rel=5e-4
            ) == value(
                CoolPropWrapper.enth_mol_ig_comp.return_expression(
                    m.fs.state[0], m.fs.props.benzene, T * pyunits.K
                )
            )

    @pytest.mark.unit
    def test_entr_mol_liq(self, m):
        assert isinstance(m.fs.props.benzene.entr_mol_liq_comp_coeff_A0, Var)
        assert isinstance(m.fs.props.benzene.entr_mol_liq_comp_coeff_A1, Var)
        assert isinstance(m.fs.props.benzene.entr_mol_liq_comp_coeff_A2, Var)
        assert isinstance(m.fs.props.benzene.entr_mol_liq_comp_coeff_A3, Var)
        assert isinstance(m.fs.props.benzene.entr_mol_liq_comp_coeff_A4, Var)
        assert isinstance(m.fs.props.benzene.entr_mol_liq_comp_coeff_A5, Var)
        assert isinstance(m.fs.props.benzene.entr_mol_liq_comp_coeff_A6, Var)
        assert isinstance(m.fs.props.benzene.entr_mol_liq_comp_coeff_A7, Var)

        assert isinstance(m.fs.props.benzene.entr_mol_liq_comp_coeff_B0, Var)
        assert isinstance(m.fs.props.benzene.entr_mol_liq_comp_coeff_B1, Var)

        assert m.fs.props.benzene.entr_mol_liq_comp_coeff_A0.value == pytest.approx(
            -367.28736141903704, rel=1e-10
        )
        assert m.fs.props.benzene.entr_mol_liq_comp_coeff_A0.fixed
        assert m.fs.props.benzene.entr_mol_liq_comp_coeff_A1.value == pytest.approx(
            1.9742974534821924, rel=1e-10
        )
        assert m.fs.props.benzene.entr_mol_liq_comp_coeff_A1.fixed
        assert m.fs.props.benzene.entr_mol_liq_comp_coeff_A2.value == pytest.approx(
            -0.0044315658996985936, rel=1e-10
        )
        assert m.fs.props.benzene.entr_mol_liq_comp_coeff_A2.fixed
        assert m.fs.props.benzene.entr_mol_liq_comp_coeff_A3.value == pytest.approx(
            1.3891733323299664e-06, rel=1e-10
        )
        assert m.fs.props.benzene.entr_mol_liq_comp_coeff_A3.fixed
        assert m.fs.props.benzene.entr_mol_liq_comp_coeff_A4.value == pytest.approx(
            2.2861650201840276e-08, rel=1e-10
        )
        assert m.fs.props.benzene.entr_mol_liq_comp_coeff_A4.fixed
        assert m.fs.props.benzene.entr_mol_liq_comp_coeff_A5.value == pytest.approx(
            -6.533821280556607e-11, rel=1e-10
        )
        assert m.fs.props.benzene.entr_mol_liq_comp_coeff_A5.fixed
        assert m.fs.props.benzene.entr_mol_liq_comp_coeff_A6.value == pytest.approx(
            7.560967575590022e-14, rel=1e-10
        )
        assert m.fs.props.benzene.entr_mol_liq_comp_coeff_A6.fixed
        assert m.fs.props.benzene.entr_mol_liq_comp_coeff_A7.value == pytest.approx(
            -3.327152569338899e-17, rel=1e-10
        )
        assert m.fs.props.benzene.entr_mol_liq_comp_coeff_A7.fixed

        assert m.fs.props.benzene.entr_mol_liq_comp_coeff_B0.value == pytest.approx(
            1, rel=1e-10
        )
        assert m.fs.props.benzene.entr_mol_liq_comp_coeff_B0.fixed
        assert m.fs.props.benzene.entr_mol_liq_comp_coeff_B1.value == pytest.approx(
            -0.0017637184112409214, rel=1e-10
        )
        assert m.fs.props.benzene.entr_mol_liq_comp_coeff_B1.fixed

        assert m.fs.props.benzene.entr_mol_liq_comp_anchor.value == pytest.approx(
            108.66827371730366, rel=1e-10
        )
        assert m.fs.props.benzene.entr_mol_liq_comp_anchor.fixed

        for T in range(300, 401, 10):
            m.fs.state[0].temperature.fix(T)
            assert pytest.approx(
                CoolProp.PropsSI("Smolar", "T", T, "Q", 0, "benzene"), rel=5e-4
            ) == value(
                CoolPropWrapper.entr_mol_liq_comp.return_expression(
                    m.fs.state[0], m.fs.props.benzene, T * pyunits.K
                )
            )

    @pytest.mark.unit
    def test_entr_mol_ig(self, m):
        assert isinstance(m.fs.props.benzene.entr_mol_ig_comp_coeff_A0, Var)
        assert isinstance(m.fs.props.benzene.entr_mol_ig_comp_coeff_A1, Var)
        assert isinstance(m.fs.props.benzene.entr_mol_ig_comp_coeff_A2, Var)
        assert isinstance(m.fs.props.benzene.entr_mol_ig_comp_coeff_A3, Var)
        assert isinstance(m.fs.props.benzene.entr_mol_ig_comp_coeff_A4, Var)
        assert isinstance(m.fs.props.benzene.entr_mol_ig_comp_coeff_A5, Var)
        assert isinstance(m.fs.props.benzene.entr_mol_ig_comp_coeff_A6, Var)
        assert isinstance(m.fs.props.benzene.entr_mol_ig_comp_coeff_A7, Var)

        assert isinstance(m.fs.props.benzene.entr_mol_ig_comp_coeff_B0, Var)
        assert isinstance(m.fs.props.benzene.entr_mol_ig_comp_coeff_B1, Var)

        assert m.fs.props.benzene.entr_mol_ig_comp_coeff_A0.value == pytest.approx(
            805.7903359684354, rel=1e-10
        )
        assert m.fs.props.benzene.entr_mol_ig_comp_coeff_A0.fixed
        assert m.fs.props.benzene.entr_mol_ig_comp_coeff_A1.value == pytest.approx(
            -7.054023595110686, rel=1e-10
        )
        assert m.fs.props.benzene.entr_mol_ig_comp_coeff_A1.fixed
        assert m.fs.props.benzene.entr_mol_ig_comp_coeff_A2.value == pytest.approx(
            0.026948845504297693, rel=1e-10
        )
        assert m.fs.props.benzene.entr_mol_ig_comp_coeff_A2.fixed
        assert m.fs.props.benzene.entr_mol_ig_comp_coeff_A3.value == pytest.approx(
            -4.872969272569453e-05, rel=1e-10
        )
        assert m.fs.props.benzene.entr_mol_ig_comp_coeff_A3.fixed
        assert m.fs.props.benzene.entr_mol_ig_comp_coeff_A4.value == pytest.approx(
            1.4263950361676005e-08, rel=1e-10
        )
        assert m.fs.props.benzene.entr_mol_ig_comp_coeff_A4.fixed
        assert m.fs.props.benzene.entr_mol_ig_comp_coeff_A5.value == pytest.approx(
            9.666423804354728e-11, rel=1e-10
        )
        assert m.fs.props.benzene.entr_mol_ig_comp_coeff_A5.fixed
        assert m.fs.props.benzene.entr_mol_ig_comp_coeff_A6.value == pytest.approx(
            -1.5555968001310246e-13, rel=1e-10
        )
        assert m.fs.props.benzene.entr_mol_ig_comp_coeff_A6.fixed
        assert m.fs.props.benzene.entr_mol_ig_comp_coeff_A7.value == pytest.approx(
            7.656482336490157e-17, rel=1e-10
        )
        assert m.fs.props.benzene.entr_mol_ig_comp_coeff_A7.fixed

        assert m.fs.props.benzene.entr_mol_ig_comp_coeff_B0.value == pytest.approx(
            1, rel=1e-10
        )
        assert m.fs.props.benzene.entr_mol_ig_comp_coeff_B0.fixed
        assert m.fs.props.benzene.entr_mol_ig_comp_coeff_B1.value == pytest.approx(
            -0.0017620273519612887, rel=1e-10
        )
        assert m.fs.props.benzene.entr_mol_ig_comp_coeff_B1.fixed

        for T in range(300, 401, 10):
            m.fs.state[0].temperature.fix(T)
            assert pytest.approx(
                CoolProp.PropsSI("Smolar", "T", T, "Q", 1, "benzene"), rel=5e-4
            ) == value(
                CoolPropWrapper.entr_mol_ig_comp.return_expression(
                    m.fs.state[0], m.fs.props.benzene, T * pyunits.K
                )
            )

    @pytest.mark.unit
    def test_psat(self, m):
        assert isinstance(m.fs.props.benzene.pressure_sat_coeff_n1, Var)
        assert isinstance(m.fs.props.benzene.pressure_sat_coeff_n2, Var)
        assert isinstance(m.fs.props.benzene.pressure_sat_coeff_n3, Var)
        assert isinstance(m.fs.props.benzene.pressure_sat_coeff_n4, Var)
        assert isinstance(m.fs.props.benzene.pressure_sat_coeff_n5, Var)
        assert isinstance(m.fs.props.benzene.pressure_sat_coeff_n6, Var)

        assert isinstance(m.fs.props.benzene.pressure_sat_coeff_t1, Var)
        assert isinstance(m.fs.props.benzene.pressure_sat_coeff_t2, Var)
        assert isinstance(m.fs.props.benzene.pressure_sat_coeff_t3, Var)
        assert isinstance(m.fs.props.benzene.pressure_sat_coeff_t4, Var)
        assert isinstance(m.fs.props.benzene.pressure_sat_coeff_t5, Var)
        assert isinstance(m.fs.props.benzene.pressure_sat_coeff_t6, Var)

        assert m.fs.props.benzene.pressure_sat_coeff_n1.value == pytest.approx(
            0.005561906558935796, rel=1e-10
        )
        assert m.fs.props.benzene.pressure_sat_coeff_n1.fixed
        assert m.fs.props.benzene.pressure_sat_coeff_n2.value == pytest.approx(
            -0.08662136922915314, rel=1e-10
        )
        assert m.fs.props.benzene.pressure_sat_coeff_n2.fixed
        assert m.fs.props.benzene.pressure_sat_coeff_n3.value == pytest.approx(
            -6.964182734154488, rel=1e-10
        )
        assert m.fs.props.benzene.pressure_sat_coeff_n3.fixed
        assert m.fs.props.benzene.pressure_sat_coeff_n4.value == pytest.approx(
            1.1249288132278856, rel=1e-10
        )
        assert m.fs.props.benzene.pressure_sat_coeff_n4.fixed
        assert m.fs.props.benzene.pressure_sat_coeff_n5.value == pytest.approx(
            -3.961859460597414, rel=1e-10
        )
        assert m.fs.props.benzene.pressure_sat_coeff_n5.fixed
        assert m.fs.props.benzene.pressure_sat_coeff_n6.value == pytest.approx(
            -13.106880507410812, rel=1e-10
        )
        assert m.fs.props.benzene.pressure_sat_coeff_n6.fixed

        assert m.fs.props.benzene.pressure_sat_coeff_t1.value == pytest.approx(
            0.037, rel=1e-10
        )
        assert m.fs.props.benzene.pressure_sat_coeff_t1.fixed
        assert m.fs.props.benzene.pressure_sat_coeff_t2.value == pytest.approx(
            0.505, rel=1e-10
        )
        assert m.fs.props.benzene.pressure_sat_coeff_t2.fixed
        assert m.fs.props.benzene.pressure_sat_coeff_t3.value == pytest.approx(
            1.014, rel=1e-10
        )
        assert m.fs.props.benzene.pressure_sat_coeff_t3.fixed
        assert m.fs.props.benzene.pressure_sat_coeff_t4.value == pytest.approx(
            1.469, rel=1e-10
        )
        assert m.fs.props.benzene.pressure_sat_coeff_t4.fixed
        assert m.fs.props.benzene.pressure_sat_coeff_t5.value == pytest.approx(
            3.711, rel=1e-10
        )
        assert m.fs.props.benzene.pressure_sat_coeff_t5.fixed
        assert m.fs.props.benzene.pressure_sat_coeff_t6.value == pytest.approx(
            12.647, rel=1e-10
        )
        assert m.fs.props.benzene.pressure_sat_coeff_t6.fixed

        for T in range(300, 401, 10):
            m.fs.state[0].temperature.fix(T)
            assert pytest.approx(
                CoolProp.PropsSI("P", "T", T, "Q", 0.5, "benzene"), rel=5e-4
            ) == value(m.fs.state[0].pressure_sat_comp["benzene"])


class TestVerifyExcessLiq(object):
    @pytest.fixture(scope="class")
    def m(self):
        # Clear cached components to ensure clean slate
        CoolPropWrapper.flush_cached_components()

        m = ConcreteModel()

        m.fs = FlowsheetBlock(dynamic=False)

        configuration = {
            # Specifying components
            "components": {
                "benzene": {
                    "type": Component,
                    "elemental_composition": {"H": 6, "C": 6},
                    "dens_mol_liq_comp": CoolPropWrapper,
                    "enth_mol_ig_comp": Constant,
                    "entr_mol_ig_comp": Constant,
                    "parameter_data": {
                        "mw": CoolPropWrapper,
                        "dens_mol_crit": CoolPropWrapper,
                        "pressure_crit": CoolPropWrapper,
                        "temperature_crit": CoolPropWrapper,
                        "omega": CoolPropWrapper,
                        "cp_mol_ig_comp_coeff": 0,
                        "enth_mol_form_ig_comp_ref": 0,
                        "entr_mol_form_ig_comp_ref": 0,
                    },
                }
            },
            # Specifying phases
            "phases": {
                "Liq": {
                    "type": LiquidPhase,
                    "equation_of_state": Cubic,
                    "equation_of_state_options": {"type": CubicType.PR},
                }
            },
            # Set base units of measurement
            "base_units": {
                "time": pyunits.s,
                "length": pyunits.m,
                "mass": pyunits.kg,
                "amount": pyunits.mol,
                "temperature": pyunits.K,
            },
            # Specifying state definition
            "state_definition": FTPx,
            "state_bounds": {
                "flow_mol": (0, 100, 1000, pyunits.mol / pyunits.s),
                "temperature": (273.15, 300, 500, pyunits.K),
                "pressure": (5e4, 1e5, 1e6, pyunits.Pa),
            },
            "pressure_ref": (101325, pyunits.Pa),
            "temperature_ref": (298.15, pyunits.K),
            "parameter_data": {"PR_kappa": {("benzene", "benzene"): 0.000}},
        }

        m.fs.props = GenericParameterBlock(**configuration)

        m.fs.state = m.fs.props.build_state_block([0], defined_state=True)

        m.fs.state[0].flow_mol.fix(1)
        m.fs.state[0].mole_frac_comp["benzene"].fix(1)

        return m

    @pytest.mark.integration
    def test_cubic_liquid(self, m):
        for P in range(1, 11):
            Td = CoolProp.PropsSI("T", "P", P * 1e5, "Q", 0.5, "PR::benzene")
            Tmin = CoolProp.PropsSI("TMIN", "benzene")

            m.fs.state[0].pressure.fix(P * 1e5)
            m.fs.state[0].temperature.fix(Tmin)

            m.fs.state.initialize()

            for T in arange(Tmin, Td, 10):
                print(P, T)
                m.fs.state[0].temperature.fix(T)

                results = solver.solve(m.fs)

                assert (
                    results.solver.termination_condition == TerminationCondition.optimal
                )
                assert results.solver.status == SolverStatus.ok

                # Check results
                assert pytest.approx(
                    CoolProp.PropsSI("Z", "T", T, "P", P * 1e5, "PR::benzene"), rel=1e-8
                ) == value(m.fs.state[0].compress_fact_phase["Liq"])

                assert pytest.approx(
                    CoolProp.PropsSI("DMOLAR", "T", T, "P", P * 1e5, "PR::benzene"),
                    rel=1e-6,
                ) == value(m.fs.state[0].dens_mol_phase["Liq"])

                assert pytest.approx(
                    CoolProp.PropsSI(
                        "HMOLAR_RESIDUAL", "T", T, "P", P * 1e5, "PR::benzene"
                    ),
                    rel=1e-6,
                ) == value(m.fs.state[0].enth_mol_phase["Liq"])

    @pytest.mark.integration
    def test_cubic_liquid_entr(self, m):
        Td = CoolProp.PropsSI("T", "P", 101325, "Q", 0.5, "PR::benzene")
        Tmin = CoolProp.PropsSI("TMIN", "benzene")

        for T in arange(Tmin, Td, 10):
            m.fs.state[0].pressure.fix(101325)
            m.fs.state[0].temperature.fix(T)

            m.fs.state.initialize()

            S0_CP = CoolProp.PropsSI("SMOLAR", "T", T, "P", 101325, "PR::benzene")
            S0_I = value(m.fs.state[0].entr_mol_phase["Liq"])

            for P in range(1, 11):
                print(T, P)
                m.fs.state[0].pressure.fix(P * 1e5)

                results = solver.solve(m.fs)

                assert (
                    results.solver.termination_condition == TerminationCondition.optimal
                )
                assert results.solver.status == SolverStatus.ok

                assert pytest.approx(
                    CoolProp.PropsSI("SMOLAR", "T", T, "P", P * 1e5, "PR::benzene")
                    - S0_CP,
                    rel=1e-4,
                ) == value(m.fs.state[0].entr_mol_phase["Liq"] - S0_I)


class TestVerifyExcessVap(object):
    @pytest.fixture(scope="class")
    def m(self):
        # Clear cached components to ensure clean slate
        CoolPropWrapper.flush_cached_components()

        m = ConcreteModel()

        m.fs = FlowsheetBlock(dynamic=False)

        # NBP reference state: h=0, s=0 for saturated liquid at 1 atmosphere
        CoolProp.set_reference_state("benzene", "NBP")

        # Get Tsat for benezene at 1 atm
        Tref = CoolProp.PropsSI("T", "P", 101325, "Q", 0.5, "PR::benzene")

        configuration = {
            # Specifying components
            "components": {
                "benzene": {
                    "type": Component,
                    "elemental_composition": {"H": 6, "C": 6},
                    "dens_mol_liq_comp": CoolPropWrapper,
                    "enth_mol_ig_comp": Constant,
                    "entr_mol_ig_comp": Constant,
                    "parameter_data": {
                        "mw": CoolPropWrapper,
                        "dens_mol_crit": CoolPropWrapper,
                        "pressure_crit": CoolPropWrapper,
                        "temperature_crit": CoolPropWrapper,
                        "omega": CoolPropWrapper,
                        "cp_mol_ig_comp_coeff": 0,
                        "enth_mol_form_ig_comp_ref": 0,
                        "entr_mol_form_ig_comp_ref": 0,
                    },
                }
            },
            # Specifying phases
            "phases": {
                "Vap": {
                    "type": VaporPhase,
                    "equation_of_state": Cubic,
                    "equation_of_state_options": {"type": CubicType.PR},
                }
            },
            # Set base units of measurement
            "base_units": {
                "time": pyunits.s,
                "length": pyunits.m,
                "mass": pyunits.kg,
                "amount": pyunits.mol,
                "temperature": pyunits.K,
            },
            # Specifying state definition
            "state_definition": FTPx,
            "state_bounds": {
                "flow_mol": (0, 100, 1000, pyunits.mol / pyunits.s),
                "temperature": (273.15, 300, 730, pyunits.K),
                "pressure": (5e4, 1e5, 1e6, pyunits.Pa),
            },
            "pressure_ref": (101325, pyunits.Pa),
            "temperature_ref": (Tref, pyunits.K),
            "parameter_data": {"PR_kappa": {("benzene", "benzene"): 0.000}},
        }

        m.fs.props = GenericParameterBlock(**configuration)

        m.fs.state = m.fs.props.build_state_block([0], defined_state=True)

        m.fs.state[0].flow_mol.fix(1)
        m.fs.state[0].mole_frac_comp["benzene"].fix(1)

        return m

    @pytest.mark.integration
    def test_cubic_vapor(self, m):
        for P in range(1, 11):
            Tb = CoolProp.PropsSI("T", "P", P * 1e5, "Q", 0.5, "PR::benzene")
            Tmax = CoolProp.PropsSI("TMAX", "benzene")

            m.fs.state[0].pressure.fix(P * 1e5)
            m.fs.state[0].temperature.fix(Tb + 1)

            m.fs.state.initialize()

            for T in arange(Tb + 1, Tmax, 10):
                m.fs.state[0].temperature.fix(T)

                results = solver.solve(m.fs)

                assert (
                    results.solver.termination_condition == TerminationCondition.optimal
                )
                assert results.solver.status == SolverStatus.ok

                # Check results
                assert pytest.approx(
                    CoolProp.PropsSI("Z", "T", T, "P", P * 1e5, "PR::benzene"), rel=1e-8
                ) == value(m.fs.state[0].compress_fact_phase["Vap"])

                assert pytest.approx(
                    CoolProp.PropsSI("DMOLAR", "T", T, "P", P * 1e5, "PR::benzene"),
                    rel=1e-6,
                ) == value(m.fs.state[0].dens_mol_phase["Vap"])

                assert pytest.approx(
                    CoolProp.PropsSI(
                        "HMOLAR_RESIDUAL", "T", T, "P", P * 1e5, "PR::benzene"
                    ),
                    rel=1e-6,
                ) == value(m.fs.state[0].enth_mol_phase["Vap"])

    @pytest.mark.integration
    def test_cubic_vapor_entr(self, m):
        Tb = CoolProp.PropsSI("T", "P", 1e6, "Q", 0.5, "PR::benzene")
        Tmax = CoolProp.PropsSI("TMAX", "benzene")

        for T in arange(Tb + 1, Tmax, 10):
            m.fs.state[0].pressure.fix(101325)
            m.fs.state[0].temperature.fix(T)

            m.fs.state.initialize()

            S0_CP = CoolProp.PropsSI("SMOLAR", "T", T, "P", 101325, "PR::benzene")
            S0_I = value(m.fs.state[0].entr_mol_phase["Vap"])

            for P in range(1, 11):
                print(T, P)
                m.fs.state[0].pressure.fix(P * 1e5)

                results = solver.solve(m.fs)

                assert (
                    results.solver.termination_condition == TerminationCondition.optimal
                )
                assert results.solver.status == SolverStatus.ok

                assert pytest.approx(
                    CoolProp.PropsSI("SMOLAR", "T", T, "P", P * 1e5, "PR::benzene")
                    - S0_CP,
                    rel=1e-4,
                ) == value(m.fs.state[0].entr_mol_phase["Vap"] - S0_I)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Author: Andrew Lee
"""

import pytest

from numpy import logspace

from pyomo.util.check_units import assert_units_consistent
from pyomo.environ import (
    check_optimal_termination,
    ConcreteModel,
    Objective,
    units as pyunits,
    value,
)

from idaes.core import FlowsheetBlock
from idaes.models.properties.modular_properties.eos.ceos import cubic_roots_available
from idaes.models.properties.modular_properties.base.generic_property import (
    GenericParameterBlock,
)
from idaes.core.solvers import get_solver
import idaes.core.util.scaling as iscale
from idaes.models.properties.tests.test_harness import PropertyTestHarness
from idaes.core import LiquidPhase, VaporPhase, Component
from idaes.models.properties.modular_properties.state_definitions import FTPx
from idaes.models.properties.modular_properties.eos.ceos import Cubic, CubicType
from idaes.models.properties.modular_properties.phase_equil import SmoothVLE
from idaes.models.properties.modular_properties.phase_equil.bubble_dew import (
    LogBubbleDew,
)
from idaes.models.properties.modular_properties.phase_equil.forms import log_fugacity
from idaes.models.properties.modular_properties.pure import RPP4

import idaes.logger as idaeslog

SOUT = idaeslog.INFO

# Set module level pyest marker
pytestmark = pytest.mark.cubic_root


# -----------------------------------------------------------------------------
# Get default solver for testing
# Limit iterations to make sure sweeps aren't getting out of hand
solver = get_solver(solver="ipopt_v2", solver_options={"max_iter": 50})

# ---------------------------------------------------------------------
# Configuration dictionary for an ideal Benzene-Toluene system

# Data Sources:
# [1] The Properties of Gases and Liquids (1987)
#     4th edition, Chemical Engineering Series - Robert C. Reid
# [3] Engineering Toolbox, https://www.engineeringtoolbox.com
#     Retrieved 1st December, 2019

configuration = {
    # Specifying components
    "components": {
        "benzene": {
            "type": Component,
            "enth_mol_ig_comp": RPP4,
            "entr_mol_ig_comp": RPP4,
            "pressure_sat_comp": RPP4,
            "phase_equilibrium_form": {("Vap", "Liq"): log_fugacity},
            "parameter_data": {
                "mw": (78.1136e-3, pyunits.kg / pyunits.mol),  # [1]
                "pressure_crit": (48.9e5, pyunits.Pa),  # [1]
                "temperature_crit": (562.2, pyunits.K),  # [1]
                "omega": 0.212,  # [1]
                "cp_mol_ig_comp_coeff": {
                    "A": (-3.392e1, pyunits.J / pyunits.mol / pyunits.K),  # [1]
                    "B": (4.739e-1, pyunits.J / pyunits.mol / pyunits.K**2),
                    "C": (-3.017e-4, pyunits.J / pyunits.mol / pyunits.K**3),
                    "D": (7.130e-8, pyunits.J / pyunits.mol / pyunits.K**4),
                },
                "enth_mol_form_vap_comp_ref": (82.9e3, pyunits.J / pyunits.mol),  # [3]
                "entr_mol_form_vap_comp_ref": (
                    -269,
                    pyunits.J / pyunits.mol / pyunits.K,
                ),  # [3]
                "pressure_sat_comp_coeff": {
                    "A": (-6.98273, None),  # [1]
                    "B": (1.33213, None),
                    "C": (-2.62863, None),
                    "D": (-3.33399, None),
                },
            },
        },
        "toluene": {
            "type": Component,
            "enth_mol_ig_comp": RPP4,
            "entr_mol_ig_comp": RPP4,
            "pressure_sat_comp": RPP4,
            "phase_equilibrium_form": {("Vap", "Liq"): log_fugacity},
            "parameter_data": {
                "mw": (92.1405e-3, pyunits.kg / pyunits.mol),  # [1]
                "pressure_crit": (41e5, pyunits.Pa),  # [1]
                "temperature_crit": (591.8, pyunits.K),  # [1]
                "omega": 0.263,  # [1]
                "cp_mol_ig_comp_coeff": {
                    "A": (-2.435e1, pyunits.J / pyunits.mol / pyunits.K),  # [1]
                    "B": (5.125e-1, pyunits.J / pyunits.mol / pyunits.K**2),
                    "C": (-2.765e-4, pyunits.J / pyunits.mol / pyunits.K**3),
                    "D": (4.911e-8, pyunits.J / pyunits.mol / pyunits.K**4),
                },
                "enth_mol_form_vap_comp_ref": (50.1e3, pyunits.J / pyunits.mol),  # [3]
                "entr_mol_form_vap_comp_ref": (
                    -321,
                    pyunits.J / pyunits.mol / pyunits.K,
                ),  # [3]
                "pressure_sat_comp_coeff": {
                    "A": (-7.28607, None),  # [1]
                    "B": (1.38091, None),
                    "C": (-2.83433, None),
                    "D": (-2.79168, None),
                },
            },
        },
    },
    # Specifying phases
    "phases": {
        "Liq": {
            "type": LiquidPhase,
            "equation_of_state": Cubic,
            "equation_of_state_options": {"type": CubicType.PR},
        },
        "Vap": {
            "type": VaporPhase,
            "equation_of_state": Cubic,
            "equation_of_state_options": {"type": CubicType.PR},
        },
    },
    # Set base units of measurement
    "base_units": {
        "time": pyunits.s,
        "length": pyunits.m,
        "mass": pyunits.kg,
        "amount": pyunits.mol,
        "temperature": pyunits.K,
    },
    # Specifying state definition
    "state_definition": FTPx,
    "state_bounds": {
        "flow_mol": (0, 100, 1000, pyunits.mol / pyunits.s),
        "temperature": (273.15, 300, 500, pyunits.K),
        "pressure": (5e4, 1e5, 1e6, pyunits.Pa),
    },
    "pressure_ref": (101325, pyunits.Pa),
    "temperature_ref": (298.15, pyunits.K),
    # Defining phase equilibria
    "phases_in_equilibrium": [("Vap", "Liq")],
    "phase_equilibrium_state": {("Vap", "Liq"): SmoothVLE},
    "bubble_dew_method": LogBubbleDew,
    "parameter_data": {
        "PR_kappa": {
            ("benzene", "benzene"): 0.000,
            ("benzene", "toluene"): 0.000,
            ("toluene", "benzene"): 0.000,
            ("toluene", "toluene"): 0.000,
        }
    },
}


@pytest.mark.skipif(not cubic_roots_available(), reason="Cubic functions not available")
class TestBTPR(PropertyTestHarness):
    def configure(self):
        self.prop_pack = GenericParameterBlock
        self.param_args = configuration
        self.prop_args = {}
        self.has_density_terms = False


# -----------------------------------------------------------------------------
# Test robustness and some outputs
@pytest.mark.skipif(not cubic_roots_available(), reason="Cubic functions not available")
class TestBTExample(object):
    @pytest.fixture()
    def m(self):
        m = ConcreteModel()

        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.props = GenericParameterBlock(**configuration)

        m.fs.state = m.fs.props.build_state_block([1], defined_state=True)

        iscale.calculate_scaling_factors(m.fs.props)
        iscale.calculate_scaling_factors(m.fs.state[1])

        return m

    @pytest.mark.integration
    def test_T_sweep(self, m):
        assert_units_consistent(m)

        m.fs.obj = Objective(expr=(m.fs.state[1].temperature - 510) ** 2)
        m.fs.state[1].temperature.setub(600)

        for P in logspace(4.8, 5.9, 8):
            m.fs.obj.deactivate()

            m.fs.state[1].flow_mol.fix(100)
            m.fs.state[1].mole_frac_comp["benzene"].fix(0.5)
            m.fs.state[1].mole_frac_comp["toluene"].fix(0.5)
            m.fs.state[1].temperature.fix(300)
            m.fs.state[1].pressure.fix(P)

            m.fs.state.initialize()

            m.fs.state[1].temperature.unfix()
            m.fs.obj.activate()

            results = solver.solve(m)

            assert check_optimal_termination(results)
            assert m.fs.state[1].flow_mol_phase["Liq"].value <= 1e-5

    @pytest.mark.integration
    def test_P_sweep(self, m):
        for T in range(370, 500, 25):
            m.fs.state[1].flow_mol.fix(100)
            m.fs.state[1].mole_frac_comp["benzene"].fix(0.5)
            m.fs.state[1].mole_frac_comp["toluene"].fix(0.5)
            m.fs.state[1].temperature.fix(T)
            m.fs.state[1].pressure.fix(1e5)

            m.fs.state.initialize()

            results = solver.solve(m)

            assert check_optimal_termination(results)

            while m.fs.state[1].pressure.value <= 1e6:

                results = solver.solve(m)
                assert check_optimal_termination(results)

                m.fs.state[1].pressure.value = m.fs.state[1].pressure.value + 1e5

    @pytest.mark.component
    def test_T350_P1_x5(self, m):
        m.fs.state[1].flow_mol.fix(100)
        m.fs.state[1].mole_frac_comp["benzene"].fix(0.5)
        m.fs.state[1].mole_frac_comp["toluene"].fix(0.5)
        m.fs.state[1].temperature.fix(350)
        m.fs.state[1].pressure.fix(1e5)

        # Trigger build of enthalpy and entropy
        m.fs.state[1].enth_mol_phase
        m.fs.state[1].entr_mol_phase

        m.fs.state.initialize(outlvl=SOUT)

        from idaes.core.util import DiagnosticsToolbox

        dt = DiagnosticsToolbox(m.fs.state[1])
        dt.report_structural_issues()
        dt.display_overconstrained_set()

        results = solver.solve(m, tee=True)

        # Check for optimal solution
        assert check_optimal_termination(results)

        assert pytest.approx(value(m.fs.state[1]._teq[("Vap", "Liq")]), abs=1e-1) == 365
        assert 0.0035346 == pytest.approx(
            value(m.fs.state[1].compress_fact_phase["Liq"]), 1e-5
        )
        assert 0.966749 == pytest.approx(
            value(m.fs.state[1].compress_fact_phase["Vap"]), 1e-5
        )
        assert (
            pytest.approx(
                value(m.fs.state[1].fug_coeff_phase_comp["Liq", "benzene"]), 1e-5
            )
            == 0.894676
        )
        assert (
            pytest.approx(
                value(m.fs.state[1].fug_coeff_phase_comp["Liq", "toluene"]), 1e-5
            )
            == 0.347566
        )
        assert (
            pytest.approx(
                value(m.fs.state[1].fug_coeff_phase_comp["Vap", "benzene"]), 1e-5
            )
            == 0.971072
        )
        assert (
            pytest.approx(
                value(m.fs.state[1].fug_coeff_phase_comp["Vap", "toluene"]), 1e-5
            )
            == 0.959791
        )

        assert (
            pytest.approx(
                value(m.fs.state[1].mole_frac_phase_comp["Liq", "benzene"]), 1e-5
            )
            == 0.5
        )
        assert (
            pytest.approx(
                value(m.fs.state[1].mole_frac_phase_comp["Liq", "toluene"]), 1e-5
            )
            == 0.5
        )
        assert (
            pytest.approx(
                value(m.fs.state[1].mole_frac_phase_comp["Vap", "benzene"]), 1e-5
            )
            == 0.70584
        )
        assert (
            pytest.approx(
                value(m.fs.state[1].mole_frac_phase_comp["Vap", "toluene"]), 1e-5
            )
            == 0.29416
        )

        assert (
            pytest.approx(value(m.fs.state[1].enth_mol_phase["Liq"]), 1e-5) == 38942.8
        )
        assert (
            pytest.approx(value(m.fs.state[1].enth_mol_phase["Vap"]), 1e-5) == 78048.7
        )
        assert (
            pytest.approx(value(m.fs.state[1].entr_mol_phase["Liq"]), 1e-5) == -361.794
        )
        assert (
            pytest.approx(value(m.fs.state[1].entr_mol_phase["Vap"]), 1e-5) == -264.0181
        )

    @pytest.mark.component
    def test_T350_P5_x5(self, m):
        m.fs.state[1].flow_mol.fix(100)
        m.fs.state[1].mole_frac_comp["benzene"].fix(0.5)
        m.fs.state[1].mole_frac_comp["toluene"].fix(0.5)
        m.fs.state[1].temperature.fix(350)
        m.fs.state[1].pressure.fix(5e5)

        # Trigger build of enthalpy and entropy
        m.fs.state[1].enth_mol_phase
        m.fs.state[1].entr_mol_phase

        m.fs.state.initialize(outlvl=SOUT)

        results = solver.solve(m)

        # Check for optimal solution
        assert check_optimal_termination(results)

        assert pytest.approx(value(m.fs.state[1]._teq[("Vap", "Liq")]), 1e-5) == 431.47
        assert (
            pytest.approx(value(m.fs.state[1].compress_fact_phase["Liq"]), 1e-5)
            == 0.01766
        )
        assert (
            pytest.approx(value(m.fs.state[1].compress_fact_phase["Vap"]), 1e-5)
            == 0.80245
        )
        assert (
            pytest.approx(
                value(m.fs.state[1].fug_coeff_phase_comp["Liq", "benzene"]), 1e-5
            )
            == 0.181229
        )
        assert (
            pytest.approx(
                value(m.fs.state[1].fug_coeff_phase_comp["Liq", "toluene"]), 1e-5
            )
            == 0.070601
        )
        assert (
            pytest.approx(
                value(m.fs.state[1].fug_coeff_phase_comp["Vap", "benzene"]), 1e-5
            )
            == 0.856523
        )
        assert (
            pytest.approx(
                value(m.fs.state[1].fug_coeff_phase_comp["Vap", "toluene"]), 1e-5
            )
            == 0.799237
        )

        assert (
            pytest.approx(
                value(m.fs.state[1].mole_frac_phase_comp["Liq", "benzene"]), 1e-5
            )
            == 0.5
        )
        assert (
            pytest.approx(
                value(m.fs.state[1].mole_frac_phase_comp["Liq", "toluene"]), 1e-5
            )
            == 0.5
        )
        assert (
            pytest.approx(
                value(m.fs.state[1].mole_frac_phase_comp["Vap", "benzene"]), 1e-5
            )
            == 0.65415
        )
        assert (
            pytest.approx(
                value(m.fs.state[1].mole_frac_phase_comp["Vap", "toluene"]), 1e-5
            )
            == 0.34585
        )

        assert (
            pytest.approx(value(m.fs.state[1].enth_mol_phase["Liq"]), 1e-5) == 38966.9
        )
        assert (
            pytest.approx(value(m.fs.state[1].enth_mol_phase["Vap"]), 1e-5) == 75150.7
        )
        assert (
            pytest.approx(value(m.fs.state[1].entr_mol_phase["Liq"]), 1e-5) == -361.8433
        )
        assert (
            pytest.approx(value(m.fs.state[1].entr_mol_phase["Vap"]), 1e-5) == -281.9703
        )

    @pytest.mark.component
    def test_T450_P1_x5(self, m):
        m.fs.state[1].flow_mol.fix(100)
        m.fs.state[1].mole_frac_comp["benzene"].fix(0.5)
        m.fs.state[1].mole_frac_comp["toluene"].fix(0.5)
        m.fs.state[1].temperature.fix(450)
        m.fs.state[1].pressure.fix(1e5)

        # Trigger build of enthalpy and entropy
        m.fs.state[1].enth_mol_phase
        m.fs.state[1].entr_mol_phase

        m.fs.state.initialize(outlvl=SOUT)

        results = solver.solve(m)

        # Check for optimal solution
        assert check_optimal_termination(results)

        assert pytest.approx(value(m.fs.state[1]._teq[("Vap", "Liq")]), 1e-5) == 371.4
        assert 0.0033583 == pytest.approx(
            value(m.fs.state[1].compress_fact_phase["Liq"]), 1e-5
        )
        assert 0.9821368 == pytest.approx(
            value(m.fs.state[1].compress_fact_phase["Vap"]), 1e-5
        )
        assert (
            pytest.approx(
                value(m.fs.state[1].fug_coeff_phase_comp["Liq", "benzene"]), 1e-5
            )
            == 8.069323
        )
        assert (
            pytest.approx(
                value(m.fs.state[1].fug_coeff_phase_comp["Liq", "toluene"]), 1e-5
            )
            == 4.304955
        )
        assert (
            pytest.approx(
                value(m.fs.state[1].fug_coeff_phase_comp["Vap", "benzene"]), 1e-5
            )
            == 0.985365
        )
        assert (
            pytest.approx(
                value(m.fs.state[1].fug_coeff_phase_comp["Vap", "toluene"]), 1e-5
            )
            == 0.979457
        )

        assert (
            pytest.approx(
                value(m.fs.state[1].mole_frac_phase_comp["Liq", "benzene"]), 1e-5
            )
            == 0.29861
        )
        assert (
            pytest.approx(
                value(m.fs.state[1].mole_frac_phase_comp["Liq", "toluene"]), 1e-5
            )
            == 0.70139
        )
        assert (
            pytest.approx(
                value(m.fs.state[1].mole_frac_phase_comp["Vap", "benzene"]), 1e-5
            )
            == 0.5
        )
        assert (
            pytest.approx(
                value(m.fs.state[1].mole_frac_phase_comp["Vap", "toluene"]), 1e-5
            )
            == 0.5
        )

        assert (
            pytest.approx(value(m.fs.state[1].enth_mol_phase["Liq"]), 1e-5) == 49441.2
        )
        assert (
            pytest.approx(value(m.fs.state[1].enth_mol_phase["Vap"]), 1e-5) == 84175.1
        )
        assert (
            pytest.approx(value(m.fs.state[1].entr_mol_phase["Liq"]), 1e-5) == -328.766
        )
        assert (
            pytest.approx(value(m.fs.state[1].entr_mol_phase["Vap"]), 1e-5) == -241.622
        )

    @pytest.mark.component
    def test_T450_P5_x5(self, m):
        m.fs.state[1].flow_mol.fix(100)
        m.fs.state[1].mole_frac_comp["benzene"].fix(0.5)
        m.fs.state[1].mole_frac_comp["toluene"].fix(0.5)
        m.fs.state[1].temperature.fix(450)
        m.fs.state[1].pressure.fix(5e5)

        # Trigger build of enthalpy and entropy
        m.fs.state[1].enth_mol_phase
        m.fs.state[1].entr_mol_phase

        m.fs.state.initialize(outlvl=SOUT)

        results = solver.solve(m)

        # Check for optimal solution
        assert check_optimal_termination(results)

        assert pytest.approx(value(m.fs.state[1]._teq[("Vap", "Liq")]), 1e-5) == 436.93
        assert 0.0166181 == pytest.approx(
            value(m.fs.state[1].compress_fact_phase["Liq"]), 1e-5
        )
        assert 0.9053766 == pytest.approx(
            value(m.fs.state[1].compress_fact_phase["Vap"]), 1e-5
        )
        assert (
            pytest.approx(
                value(m.fs.state[1].fug_coeff_phase_comp["Liq", "benzene"]), 1e-5
            )
            == 1.63308
        )
        assert (
            pytest.approx(
                value(m.fs.state[1].fug_coeff_phase_comp["Liq", "toluene"]), 1e-5
            )
            == 0.873213
        )
        assert (
            pytest.approx(
                value(m.fs.state[1].fug_coeff_phase_comp["Vap", "benzene"]), 1e-5
            )
            == 0.927534
        )
        assert (
            pytest.approx(
                value(m.fs.state[1].fug_coeff_phase_comp["Vap", "toluene"]), 1e-5
            )
            == 0.898324
        )

        assert (
            pytest.approx(
                value(m.fs.state[1].mole_frac_phase_comp["Liq", "benzene"]), 1e-5
            )
            == 0.3488737
        )
        assert (
            pytest.approx(
                value(m.fs.state[1].mole_frac_phase_comp["Liq", "toluene"]), 1e-5
            )
            == 0.6511263
        )
        assert (
            pytest.approx(
                value(m.fs.state[1].mole_frac_phase_comp["Vap", "benzene"]), 1e-5
            )
            == 0.5
        )
        assert (
            pytest.approx(
                value(m.fs.state[1].mole_frac_phase_comp["Vap", "toluene"]), 1e-5
            )
            == 0.5
        )

        assert (
            pytest.approx(value(m.fs.state[1].enth_mol_phase["Liq"]), 1e-5) == 51095.2
        )
        assert (
            pytest.approx(value(m.fs.state[1].enth_mol_phase["Vap"]), 1e-5) == 83362.3
        )
        assert (
            pytest.approx(value(m.fs.state[1].entr_mol_phase["Liq"]), 1e-5) == -326.299
        )
        assert (
            pytest.approx(value(m.fs.state[1].entr_mol_phase["Vap"]), 1e-5) == -256.198
        )

    @pytest.mark.component
    def test_T368_P1_x5(self, m):
        m.fs.state[1].flow_mol.fix(100)
        m.fs.state[1].mole_frac_comp["benzene"].fix(0.5)
        m.fs.state[1].mole_frac_comp["toluene"].fix(0.5)
        m.fs.state[1].temperature.fix(368)
        m.fs.state[1].pressure.fix(1e5)

        # Trigger build of enthalpy and entropy
        m.fs.state[1].enth_mol_phase
        m.fs.state[1].entr_mol_phase

        m.fs.state.initialize(outlvl=SOUT)

        results = solver.solve(m)

        # Check for optimal solution
        assert check_optimal_termination(results)

        assert pytest.approx(value(m.fs.state[1]._teq[("Vap", "Liq")]), 1e-5) == 368
        assert 0.003504 == pytest.approx(
            value(m.fs.state[1].compress_fact_phase["Liq"]), 1e-5
        )
        assert 0.97 == pytest.approx(
            value(m.fs.state[1].compress_fact_phase["Vap"]), 1e-5
        )
        assert (
            pytest.approx(
                value(m.fs.state[1].fug_coeff_phase_comp["Liq", "benzene"]), 1e-5
            )
            == 1.492049
        )
        assert (
            pytest.approx(
                value(m.fs.state[1].fug_coeff_phase_comp["Liq", "toluene"]), 1e-5
            )
            == 0.621563
        )
        assert (
            pytest.approx(
                value(m.fs.state[1].fug_coeff_phase_comp["Vap", "benzene"]), 1e-5
            )
            == 0.97469
        )
        assert (
            pytest.approx(
                value(m.fs.state[1].fug_coeff_phase_comp["Vap", "toluene"]), 1e-5
            )
            == 0.964642
        )

        assert (
            pytest.approx(
                value(m.fs.state[1].mole_frac_phase_comp["Liq", "benzene"]), 1e-5
            )
            == 0.4012128
        )
        assert (
            pytest.approx(
                value(m.fs.state[1].mole_frac_phase_comp["Liq", "toluene"]), 1e-5
            )
            == 0.5987872
        )

        assert (
            pytest.approx(
                value(m.fs.state[1].mole_frac_phase_comp["Vap", "benzene"]), 1e-5
            )
            == 0.6141738
        )
        assert (
            pytest.approx(
                value(m.fs.state[1].mole_frac_phase_comp["Vap", "toluene"]), 1e-5
            )
            == 0.3858262
        )

        m.fs.state[1].mole_frac_phase_comp.display()
        m.fs.state[1].enth_mol_phase_comp.display()

        assert (
            pytest.approx(value(m.fs.state[1].enth_mol_phase["Liq"]), 1e-5) == 38235.1
        )
        assert (
            pytest.approx(value(m.fs.state[1].enth_mol_phase["Vap"]), 1e-5) == 77155.4
        )
        assert (
            pytest.approx(value(m.fs.state[1].entr_mol_phase["Liq"]), 1e-5) == -359.256
        )
        assert (
            pytest.approx(value(m.fs.state[1].entr_mol_phase["Vap"]), 1e-5) == -262.348
        )

    @pytest.mark.component
    def test_T376_P1_x2(self, m):
        m.fs.state[1].flow_mol.fix(100)
        m.fs.state[1].mole_frac_comp["benzene"].fix(0.2)
        m.fs.state[1].mole_frac_comp["toluene"].fix(0.8)
        m.fs.state[1].temperature.fix(376)
        m.fs.state[1].pressure.fix(1e5)

        # Trigger build of enthalpy and entropy
        m.fs.state[1].enth_mol_phase
        m.fs.state[1].entr_mol_phase

        m.fs.state.initialize(outlvl=SOUT)

        results = solver.solve(m)

        # Check for optimal solution
        assert check_optimal_termination(results)

        assert pytest.approx(value(m.fs.state[1]._teq[("Vap", "Liq")]), 1e-5) == 376
        assert 0.00361333 == pytest.approx(
            value(m.fs.state[1].compress_fact_phase["Liq"]), 1e-5
        )
        assert 0.968749 == pytest.approx(
            value(m.fs.state[1].compress_fact_phase["Vap"]), 1e-5
        )
        assert (
            pytest.approx(
                value(m.fs.state[1].fug_coeff_phase_comp["Liq", "benzene"]), 1e-5
            )
            == 1.8394188
        )
        assert (
            pytest.approx(
                value(m.fs.state[1].fug_coeff_phase_comp["Liq", "toluene"]), 1e-5
            )
            == 0.7871415
        )
        assert (
            pytest.approx(
                value(m.fs.state[1].fug_coeff_phase_comp["Vap", "benzene"]), 1e-5
            )
            == 0.9763608
        )
        assert (
            pytest.approx(
                value(m.fs.state[1].fug_coeff_phase_comp["Vap", "toluene"]), 1e-5
            )
            == 0.9663611
        )

        assert (
            pytest.approx(
                value(m.fs.state[1].mole_frac_phase_comp["Liq", "benzene"]), 1e-5
            )
            == 0.17342
        )
        assert (
            pytest.approx(
                value(m.fs.state[1].mole_frac_phase_comp["Liq", "toluene"]), 1e-5
            )
            == 0.82658
        )
        assert (
            pytest.approx(
                value(m.fs.state[1].mole_frac_phase_comp["Vap", "benzene"]), 1e-5
            )
            == 0.3267155
        )
        assert (
            pytest.approx(
                value(m.fs.state[1].mole_frac_phase_comp["Vap", "toluene"]), 1e-5
            )
            == 0.6732845
        )

        assert (
            pytest.approx(value(m.fs.state[1].enth_mol_phase["Liq"]), 1e-5) == 31535.8
        )
        assert (
            pytest.approx(value(m.fs.state[1].enth_mol_phase["Vap"]), 1e-5) == 69175.3
        )
        assert (
            pytest.approx(value(m.fs.state[1].entr_mol_phase["Liq"]), 1e-5) == -369.033
        )
        assert (
            pytest.approx(value(m.fs.state[1].entr_mol_phase["Vap"]), 1e-5) == -273.513
        )



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests for constant pure component properties

Liquid properties are tested with with data for water
Ideal gas properties are tested with data for Air

All parameter indices based on conventions used by the source

Authors: Andres J Calderon, Andrew Lee
"""

import pytest
import types

from pyomo.environ import ConcreteModel, Block, value, units as pyunits
from pyomo.common.config import ConfigBlock
from pyomo.util.check_units import assert_units_equivalent

from idaes.models.properties.modular_properties.pure.ConstantProperties import *
from idaes.core.util.misc import add_object_reference
from idaes.core.base.property_meta import PropertyClassMetadata


@pytest.fixture()
def frame():
    m = ConcreteModel()

    # Create a dummy parameter block
    m.params = Block()

    m.params.config = ConfigBlock(implicit=True)
    m.params.config.parameter_data = {
        "dens_mol_liq_comp_coeff": (55.2046332734557, pyunits.kmol / pyunits.m**3),
        "cp_mol_liq_comp_coeff": (
            75704.2953333398,
            pyunits.J / pyunits.kmol / pyunits.K,
        ),
        "enth_mol_form_liq_comp_ref": (-285.83, pyunits.kJ / pyunits.mol),
        "entr_mol_form_liq_comp_ref": (69.95, pyunits.J / pyunits.K / pyunits.mol),
        "cp_mol_ig_comp_coeff": (29114.850, pyunits.J / pyunits.kmol / pyunits.K),
        "enth_mol_form_ig_comp_ref": (0.0, pyunits.kJ / pyunits.mol),
        "entr_mol_form_ig_comp_ref": (0.0, pyunits.J / pyunits.K / pyunits.mol),
        "dens_mol_sol_comp_coeff": (100, pyunits.kmol / pyunits.m**3),
        "cp_mol_sol_comp_coeff": (100000, pyunits.J / pyunits.kmol / pyunits.K),
        "enth_mol_form_sol_comp_ref": (-300, pyunits.kJ / pyunits.mol),
        "entr_mol_form_sol_comp_ref": (50, pyunits.J / pyunits.K / pyunits.mol),
        "visc_d_Liq_comp_coeff": (8.9e-4, pyunits.Pa * pyunits.s),
        "therm_cond_Liq_comp_coeff": (0.598, pyunits.W / pyunits.m / pyunits.K),
    }
    m.params.config.include_enthalpy_of_formation = True

    # Also need to dummy configblock on the model for the test
    m.config = ConfigBlock(implicit=True)
    m.config.include_enthalpy_of_formation = True

    m.meta_object = PropertyClassMetadata()
    m.meta_object._default_units.set_units(
        temperature=pyunits.K,
        mass=pyunits.kg,
        length=pyunits.m,
        time=pyunits.s,
        amount=pyunits.mol,
    )

    def get_metadata(self):
        return m.meta_object

    m.get_metadata = types.MethodType(get_metadata, m)
    m.params.get_metadata = types.MethodType(get_metadata, m.params)

    # Add necessary parameters to parameter block
    m.params.temperature_ref = Var(initialize=300, units=pyunits.K)

    # Create a dummy state block
    m.props = Block([1])
    add_object_reference(m.props[1], "params", m.params)

    m.props[1].temperature = Var(initialize=300, units=pyunits.K)

    return m


@pytest.mark.unit
def test_cp_mol_liq_comp(frame):
    Constant.cp_mol_liq_comp.build_parameters(frame.params)

    assert isinstance(frame.params.cp_mol_liq_comp_coeff, Var)
    assert value(frame.params.cp_mol_liq_comp_coeff) == pytest.approx(
        75704.3 / 1000, rel=1e-5
    )

    expr = Constant.cp_mol_liq_comp.return_expression(
        frame.props[1], frame.params, frame.props[1].temperature
    )
    assert value(expr) == pytest.approx(75704.3 / 1000, rel=1e-5)

    frame.props[1].temperature.value = 400
    assert value(expr) == pytest.approx(75704.3 / 1000, rel=1e-5)

    assert_units_equivalent(expr, pyunits.J / pyunits.mol / pyunits.K)


@pytest.mark.unit
def test_enth_mol_liq_comp(frame):

    Constant.enth_mol_liq_comp.build_parameters(frame.params)

    assert isinstance(frame.params.enth_mol_form_liq_comp_ref, Var)
    assert value(frame.params.enth_mol_form_liq_comp_ref) == -285.83e3

    expr = Constant.enth_mol_liq_comp.return_expression(
        frame.props[1], frame.params, frame.props[1].temperature
    )
    assert value(expr) == value(frame.params.enth_mol_form_liq_comp_ref)

    frame.props[1].temperature.value = 301
    assert value(expr) == pytest.approx(-285754.29570466664, rel=1e-3)

    frame.props[1].temperature.value = 400
    assert value(expr) == pytest.approx(-278259.570466, rel=1e-3)

    assert_units_equivalent(expr, pyunits.J / pyunits.mol)


@pytest.mark.unit
def test_enth_mol_liq_comp_no_formation(frame):
    frame.config.include_enthalpy_of_formation = False
    frame.params.config.include_enthalpy_of_formation = False

    Constant.enth_mol_liq_comp.build_parameters(frame.params)

    assert not hasattr(frame.params, "enth_mol_form_liq_comp_ref")

    expr = Constant.enth_mol_liq_comp.return_expression(
        frame.props[1], frame.params, frame.props[1].temperature
    )
    assert value(expr) == 0

    frame.props[1].temperature.value = 400
    assert value(expr) == pytest.approx(7570.42, rel=1e-3)

    assert_units_equivalent(expr, pyunits.J / pyunits.mol)


@pytest.mark.unit
def test_entr_mol_liq_comp(frame):
    Constant.entr_mol_liq_comp.build_parameters(frame.params)

    assert isinstance(frame.params.entr_mol_form_liq_comp_ref, Var)
    assert value(frame.params.entr_mol_form_liq_comp_ref) == 69.95

    expr = Constant.entr_mol_liq_comp.return_expression(
        frame.props[1], frame.params, frame.props[1].temperature
    )
    assert value(expr) == value(frame.params.entr_mol_form_liq_comp_ref)

    frame.props[1].temperature.value = 400
    assert value(expr) == pytest.approx(91.72, rel=1e-3)

    assert_units_equivalent(expr, pyunits.J / pyunits.mol / pyunits.K)


@pytest.mark.unit
def test_dens_mol_liq_comp(frame):
    Constant.dens_mol_liq_comp.build_parameters(frame.params)

    assert isinstance(frame.params.dens_mol_liq_comp_coeff, Var)
    assert value(frame.params.dens_mol_liq_comp_coeff) == pytest.approx(
        55.204e3, rel=1e-4
    )

    expr = Constant.dens_mol_liq_comp.return_expression(
        frame.props[1], frame.params, frame.props[1].temperature
    )
    assert value(expr) == pytest.approx(55.204e3, rel=1e-4)

    frame.props[1].temperature.value = 400
    assert value(expr) == pytest.approx(55.204e3, rel=1e-4)

    assert_units_equivalent(expr, pyunits.mol / pyunits.m**3)


@pytest.mark.unit
def test_cp_mol_ig_comp(frame):
    Constant.cp_mol_ig_comp.build_parameters(frame.params)

    assert isinstance(frame.params.cp_mol_ig_comp_coeff, Var)
    assert value(frame.params.cp_mol_ig_comp_coeff) == pytest.approx(
        29114.850 / 1000, rel=1e-5
    )

    expr = Constant.cp_mol_ig_comp.return_expression(
        frame.props[1], frame.params, frame.props[1].temperature
    )
    assert value(expr) == pytest.approx(29114.850 / 1000, rel=1e-5)

    frame.props[1].temperature.value = 400
    assert value(expr) == pytest.approx(29114.850 / 1000, rel=1e-5)

    assert_units_equivalent(expr, pyunits.J / pyunits.mol / pyunits.K)


@pytest.mark.unit
def test_enth_mol_ig_comp(frame):

    Constant.enth_mol_ig_comp.build_parameters(frame.params)

    assert isinstance(frame.params.enth_mol_form_ig_comp_ref, Var)
    assert value(frame.params.enth_mol_form_ig_comp_ref) == -0.0

    expr = Constant.enth_mol_ig_comp.return_expression(
        frame.props[1], frame.params, frame.props[1].temperature
    )
    assert value(expr) == value(frame.params.enth_mol_form_ig_comp_ref)

    frame.props[1].temperature.value = 301
    assert value(expr) == pytest.approx(29.114850, rel=1e-3)

    frame.props[1].temperature.value = 400
    assert value(expr) == pytest.approx(2911.485, rel=1e-3)

    assert_units_equivalent(expr, pyunits.J / pyunits.mol)


@pytest.mark.unit
def test_enth_mol_ig_comp_no_formation(frame):

    frame.config.include_enthalpy_of_formation = False
    frame.params.config.include_enthalpy_of_formation = False

    Constant.enth_mol_ig_comp.build_parameters(frame.params)

    assert not hasattr(frame.params, "enth_mol_form_ig_comp_ref")

    expr = Constant.enth_mol_ig_comp.return_expression(
        frame.props[1], frame.params, frame.props[1].temperature
    )
    assert value(expr) == 0

    frame.props[1].temperature.value = 400
    assert value(expr) == pytest.approx(2911.485, rel=1e-3)

    assert_units_equivalent(expr, pyunits.J / pyunits.mol)


@pytest.mark.unit
def test_entr_mol_ig_comp(frame):
    Constant.entr_mol_ig_comp.build_parameters(frame.params)

    assert isinstance(frame.params.entr_mol_form_ig_comp_ref, Var)
    assert value(frame.params.entr_mol_form_ig_comp_ref) == 0.0

    expr = Constant.entr_mol_ig_comp.return_expression(
        frame.props[1], frame.params, frame.props[1].temperature
    )
    assert value(expr) == value(frame.params.entr_mol_form_ig_comp_ref)

    frame.props[1].temperature.value = 400
    assert value(expr) == pytest.approx(8.37582, rel=1e-3)

    assert_units_equivalent(expr, pyunits.J / pyunits.mol / pyunits.K)


@pytest.mark.unit
def test_cp_mol_sol_comp(frame):
    Constant.cp_mol_sol_comp.build_parameters(frame.params)

    assert isinstance(frame.params.cp_mol_sol_comp_coeff, Var)
    assert value(frame.params.cp_mol_sol_comp_coeff) == pytest.approx(
        100000 / 1000, rel=1e-5
    )

    expr = Constant.cp_mol_sol_comp.return_expression(
        frame.props[1], frame.params, frame.props[1].temperature
    )
    assert value(expr) == pytest.approx(100000 / 1000, rel=1e-5)

    frame.props[1].temperature.value = 400
    assert value(expr) == pytest.approx(100000 / 1000, rel=1e-5)

    assert_units_equivalent(expr, pyunits.J / pyunits.mol / pyunits.K)


@pytest.mark.unit
def test_enth_mol_sol_comp(frame):

    Constant.enth_mol_sol_comp.build_parameters(frame.params)

    assert isinstance(frame.params.enth_mol_form_sol_comp_ref, Var)
    assert value(frame.params.enth_mol_form_sol_comp_ref) == -300e3

    expr = Constant.enth_mol_sol_comp.return_expression(
        frame.props[1], frame.params, frame.props[1].temperature
    )
    assert value(expr) == value(frame.params.enth_mol_form_sol_comp_ref)

    frame.props[1].temperature.value = 301
    assert value(expr) == pytest.approx(-299900.0, rel=1e-3)

    frame.props[1].temperature.value = 400
    assert value(expr) == pytest.approx(-290000.0, rel=1e-3)

    assert_units_equivalent(expr, pyunits.J / pyunits.mol)


@pytest.mark.unit
def test_enth_mol_sol_comp_no_formation(frame):
    frame.config.include_enthalpy_of_formation = False
    frame.params.config.include_enthalpy_of_formation = False

    Constant.enth_mol_sol_comp.build_parameters(frame.params)

    assert not hasattr(frame.params, "enth_mol_form_sol_comp_ref")

    expr = Constant.enth_mol_sol_comp.return_expression(
        frame.props[1], frame.params, frame.props[1].temperature
    )
    assert value(expr) == 0

    frame.props[1].temperature.value = 400
    assert value(expr) == pytest.approx(10000, rel=1e-3)

    assert_units_equivalent(expr, pyunits.J / pyunits.mol)


@pytest.mark.unit
def test_entr_mol_sol_comp(frame):
    Constant.entr_mol_sol_comp.build_parameters(frame.params)

    assert isinstance(frame.params.entr_mol_form_sol_comp_ref, Var)
    assert value(frame.params.entr_mol_form_sol_comp_ref) == 50

    expr = Constant.entr_mol_sol_comp.return_expression(
        frame.props[1], frame.params, frame.props[1].temperature
    )
    assert value(expr) == value(frame.params.entr_mol_form_sol_comp_ref)

    frame.props[1].temperature.value = 400
    assert value(expr) == pytest.approx(78.77, rel=1e-3)

    assert_units_equivalent(expr, pyunits.J / pyunits.mol / pyunits.K)


@pytest.mark.unit
def test_dens_mol_sol_comp(frame):
    Constant.dens_mol_sol_comp.build_parameters(frame.params)

    assert isinstance(frame.params.dens_mol_sol_comp_coeff, Var)
    assert value(frame.params.dens_mol_sol_comp_coeff) == pytest.approx(100e3, rel=1e-4)

    expr = Constant.dens_mol_sol_comp.return_expression(
        frame.props[1], frame.params, frame.props[1].temperature
    )
    assert value(expr) == pytest.approx(100e3, rel=1e-4)

    frame.props[1].temperature.value = 400
    assert value(expr) == pytest.approx(100e3, rel=1e-4)

    assert_units_equivalent(expr, pyunits.mol / pyunits.m**3)


@pytest.mark.unit
def test_visc_d_phase_comp(frame):
    Constant.visc_d_phase_comp.build_parameters(frame.params, "Liq")

    assert isinstance(frame.params.visc_d_Liq_comp_coeff, Var)
    assert value(frame.params.visc_d_Liq_comp_coeff) == pytest.approx(8.9e-4, rel=1e-5)

    expr = Constant.visc_d_phase_comp.return_expression(
        frame.props[1], frame.params, "Liq", frame.props[1].temperature
    )
    assert value(expr) == pytest.approx(8.9e-4, rel=1e-5)

    frame.props[1].temperature.value = 400
    assert value(expr) == pytest.approx(8.9e-4, rel=1e-5)

    assert_units_equivalent(expr, pyunits.Pa * pyunits.s)


@pytest.mark.unit
def test_therm_cond_liq_comp(frame):
    Constant.therm_cond_phase_comp.build_parameters(frame.params, "Liq")

    assert isinstance(frame.params.therm_cond_Liq_comp_coeff, Var)
    assert value(frame.params.therm_cond_Liq_comp_coeff) == pytest.approx(
        0.598, rel=1e-5
    )

    expr = Constant.therm_cond_phase_comp.return_expression(
        frame.props[1], frame.params, "Liq", frame.props[1].temperature
    )
    assert value(expr) == pytest.approx(0.598, rel=1e-5)

    frame.props[1].temperature.value = 400
    assert value(expr) == pytest.approx(0.598, rel=1e-5)

    assert_units_equivalent(expr, pyunits.W / pyunits.m / pyunits.K)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################

from pandas import isna
import pytest

from pyomo.environ import (
    ConcreteModel,
    Expression,
    TransformationFactory,
    Var,
    value,
    units as pyunits,
)
from pyomo.network import Arc, Port
from idaes.core import (
    FlowsheetBlock,
    StateBlock,
    StateBlockData,
    declare_process_block_class,
)
from idaes.core.util.tables import (
    arcs_to_stream_dict,
    create_stream_table_dataframe,
    create_stream_table_ui,
    stream_table_dataframe_to_string,
    generate_table,
    stream_states_dict,
)
import idaes.models.properties.examples.saponification_thermo as thermo_props
import idaes.models.properties.examples.saponification_reactions as rxn_props
from idaes.models.unit_models import CSTR, Flash
from idaes.models.unit_models.heat_exchanger_1D import HeatExchanger1D as HX1D
from idaes.core.util.testing import PhysicalParameterTestBlock
from idaes.models_extra.column_models import TrayColumn
from idaes.models_extra.column_models.condenser import CondenserType, TemperatureSpec
from idaes.models.properties.activity_coeff_models.BTX_activity_coeff_VLE import (
    BTXParameterBlock,
)


@pytest.fixture()
def m():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.thermo_params = thermo_props.SaponificationParameterBlock()
    m.fs.reaction_params = rxn_props.SaponificationReactionParameterBlock(
        property_package=m.fs.thermo_params
    )

    m.fs.tank1 = CSTR(
        property_package=m.fs.thermo_params, reaction_package=m.fs.reaction_params
    )
    m.fs.tank2 = CSTR(
        property_package=m.fs.thermo_params, reaction_package=m.fs.reaction_params
    )
    m.fs.tank_array = CSTR(
        range(3),
        property_package=m.fs.thermo_params,
        reaction_package=m.fs.reaction_params,
    )

    m.fs.stream = Arc(source=m.fs.tank1.outlet, destination=m.fs.tank2.inlet)

    def stream_array_rule(b, i):
        return (b.tank_array[i].outlet, b.tank_array[i + 1].inlet)

    m.fs.stream_array = Arc(range(2), rule=stream_array_rule)

    TransformationFactory("network.expand_arcs").apply_to(m)

    return m


@pytest.fixture()
def m_with_variable_types():
    """Flash unit model. Use '.fs' attribute to get the flowsheet."""
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    # Flash properties
    m.fs.properties = BTXParameterBlock(
        valid_phase=("Liq", "Vap"), activity_coeff_model="Ideal", state_vars="FTPz"
    )
    # Flash unit
    m.fs.flash = Flash(property_package=m.fs.properties)
    # Adding fixed and unfixed variables
    m.fs.flash.inlet.pressure.fix(3.14)
    m.fs.flash.inlet.pressure.unfix()
    m.fs.flash.inlet.temperature.fix(368)

    return m


@pytest.mark.unit
def test_create_stream_table_dataframe_from_StateBlock(m):
    d = arcs_to_stream_dict(m, descend_into=True, prepend="model")
    assert "model.stream" in d
    assert "model.stream_array" in d
    assert d["model.stream"] == m.fs.stream
    assert d["model.stream_array"] == m.fs.stream_array


@pytest.mark.unit
def test_stream_states_dict(m):
    d = stream_states_dict(arcs_to_stream_dict(m, descend_into=True))
    assert "stream" in d
    assert "stream_array[0]" in d
    assert "stream_array[1]" in d
    assert d["stream_array[0]"] == m.fs.tank_array[1].control_volume.properties_in[0]
    assert d["stream_array[1]"] == m.fs.tank_array[2].control_volume.properties_in[0]


@pytest.mark.unit
def test_create_stream_table_dataframe_from_StateBlock_2(m):
    df = create_stream_table_dataframe(arcs_to_stream_dict(m, descend_into=True))

    assert df.loc["Pressure"]["stream"] == 101325
    assert df.loc["Temperature"]["stream"] == 298.15
    assert df.loc["Volumetric Flowrate"]["stream"] == 1.0
    assert df.loc["Molar Concentration H2O"]["stream"] == 100.0
    assert df.loc["Molar Concentration NaOH"]["stream"] == 100.0
    assert df.loc["Molar Concentration EthylAcetate"]["stream"] == 100.0
    assert df.loc["Molar Concentration SodiumAcetate"]["stream"] == 100.0
    assert df.loc["Molar Concentration Ethanol"]["stream"] == 100.0

    assert df.loc["Pressure"]["stream_array[0]"] == 101325
    assert df.loc["Temperature"]["stream_array[0]"] == 298.15
    assert df.loc["Volumetric Flowrate"]["stream_array[0]"] == 1.0
    assert df.loc["Molar Concentration H2O"]["stream_array[0]"] == 100.0
    assert df.loc["Molar Concentration NaOH"]["stream_array[0]"] == 100.0
    assert df.loc["Molar Concentration EthylAcetate"]["stream_array[0]"] == 100.0
    assert df.loc["Molar Concentration SodiumAcetate"]["stream_array[0]"] == 100.0
    assert df.loc["Molar Concentration Ethanol"]["stream_array[0]"] == 100.0

    assert df.loc["Pressure"]["stream_array[1]"] == 101325
    assert df.loc["Temperature"]["stream_array[1]"] == 298.15
    assert df.loc["Volumetric Flowrate"]["stream_array[1]"] == 1.0
    assert df.loc["Molar Concentration H2O"]["stream_array[1]"] == 100.0
    assert df.loc["Molar Concentration NaOH"]["stream_array[1]"] == 100.0
    assert df.loc["Molar Concentration EthylAcetate"]["stream_array[1]"] == 100.0
    assert df.loc["Molar Concentration SodiumAcetate"]["stream_array[1]"] == 100.0
    assert df.loc["Molar Concentration Ethanol"]["stream_array[1]"] == 100.0


@pytest.mark.unit
def test_create_stream_table_dataframe_from_StateBlock_true_state(m):
    df = create_stream_table_dataframe(
        {"state": m.fs.tank1.control_volume.properties_out}, true_state=True
    )

    assert df.loc["pressure"]["state"] == 101325
    assert df.loc["temperature"]["state"] == 298.15
    assert df.loc["flow_vol"]["state"] == 1.0
    assert df.loc["conc_mol_comp H2O"]["state"] == 100.0
    assert df.loc["conc_mol_comp NaOH"]["state"] == 100.0
    assert df.loc["conc_mol_comp EthylAcetate"]["state"] == 100.0
    assert df.loc["conc_mol_comp SodiumAcetate"]["state"] == 100.0
    assert df.loc["conc_mol_comp Ethanol"]["state"] == 100.0


@pytest.mark.unit
def test_create_stream_table_dataframe_from_StateBlock_orient(m):
    df = create_stream_table_dataframe(
        {"state": m.fs.tank1.control_volume.properties_out}, orient="index"
    )

    assert df.loc["state"]["Pressure"] == 101325
    assert df.loc["state"]["Temperature"] == 298.15
    assert df.loc["state"]["Volumetric Flowrate"] == 1.0
    assert df.loc["state"]["Molar Concentration H2O"] == 100.0
    assert df.loc["state"]["Molar Concentration NaOH"] == 100.0
    assert df.loc["state"]["Molar Concentration EthylAcetate"] == 100.0
    assert df.loc["state"]["Molar Concentration SodiumAcetate"] == 100.0
    assert df.loc["state"]["Molar Concentration Ethanol"] == 100.0


@pytest.mark.unit
def test_create_stream_table_dataframe_from_StateBlock_time():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False, time_set=[3])
    m.fs.thermo_params = thermo_props.SaponificationParameterBlock()
    m.fs.reaction_params = rxn_props.SaponificationReactionParameterBlock(
        property_package=m.fs.thermo_params
    )

    m.fs.tank1 = CSTR(
        property_package=m.fs.thermo_params, reaction_package=m.fs.reaction_params
    )
    m.fs.tank2 = CSTR(
        property_package=m.fs.thermo_params, reaction_package=m.fs.reaction_params
    )

    m.fs.stream = Arc(source=m.fs.tank1.outlet, destination=m.fs.tank2.inlet)
    TransformationFactory("network.expand_arcs").apply_to(m)

    df = create_stream_table_dataframe(
        {"state": m.fs.tank1.control_volume.properties_out}, time_point=3
    )

    assert df.loc["Pressure"]["state"] == 101325
    assert df.loc["Temperature"]["state"] == 298.15
    assert df.loc["Volumetric Flowrate"]["state"] == 1.0
    assert df.loc["Molar Concentration H2O"]["state"] == 100.0
    assert df.loc["Molar Concentration NaOH"]["state"] == 100.0
    assert df.loc["Molar Concentration EthylAcetate"]["state"] == 100.0
    assert df.loc["Molar Concentration SodiumAcetate"]["state"] == 100.0
    assert df.loc["Molar Concentration Ethanol"]["state"] == 100.0


@pytest.mark.unit
def test_create_stream_table_dataframe_from_Port(m):
    df = create_stream_table_dataframe({"state": m.fs.tank1.outlet})

    assert df.loc["Pressure"]["state"] == 101325
    assert df.loc["Temperature"]["state"] == 298.15
    assert df.loc["Volumetric Flowrate"]["state"] == 1.0
    assert df.loc["Molar Concentration H2O"]["state"] == 100.0
    assert df.loc["Molar Concentration NaOH"]["state"] == 100.0
    assert df.loc["Molar Concentration EthylAcetate"]["state"] == 100.0
    assert df.loc["Molar Concentration SodiumAcetate"]["state"] == 100.0
    assert df.loc["Molar Concentration Ethanol"]["state"] == 100.0


@pytest.mark.unit
def test_create_stream_table_dataframe_from_Arc(m):
    df = create_stream_table_dataframe({"state": m.fs.stream})

    assert df.loc["Pressure"]["state"] == 101325
    assert df.loc["Temperature"]["state"] == 298.15
    assert df.loc["Volumetric Flowrate"]["state"] == 1.0
    assert df.loc["Molar Concentration H2O"]["state"] == 100.0
    assert df.loc["Molar Concentration NaOH"]["state"] == 100.0
    assert df.loc["Molar Concentration EthylAcetate"]["state"] == 100.0
    assert df.loc["Molar Concentration SodiumAcetate"]["state"] == 100.0
    assert df.loc["Molar Concentration Ethanol"]["state"] == 100.0


@pytest.mark.unit
def test_create_stream_table_ui(m_with_variable_types):
    m = m_with_variable_types

    state_name = "state"
    state_dict = {state_name: m.fs.flash.inlet}
    df = create_stream_table_ui(state_dict)

    assert df.loc["pressure"][state_name] == (3.14, "unfixed")
    assert df.loc["temperature"][state_name] == (368, "fixed")


@pytest.mark.unit
def test_create_stream_table_dataframe_wrong_type(m):
    with pytest.raises(TypeError):
        create_stream_table_dataframe({"state": m.fs.tank1})


@pytest.mark.unit
def test_create_stream_table_dataframe_ordering(m):
    state_dict = {
        "state1": m.fs.stream,
        "state3": m.fs.tank1.control_volume.properties_out,
        "state2": m.fs.tank1.outlet,
    }
    df = create_stream_table_dataframe(state_dict)

    columns = list(df)
    assert columns[0] == "Units"
    assert columns[1] == "state1"
    assert columns[2] == "state3"
    assert columns[3] == "state2"


@pytest.mark.unit
def test_stream_table_dataframe_to_string(m):
    df = create_stream_table_dataframe(
        {"state": m.fs.tank1.control_volume.properties_out}
    )

    stream_table_dataframe_to_string(df)


# -----------------------------------------------------------------------------
###
# Create a dummy StateBlock class
##
@declare_process_block_class("StateBlockForTesting", block_class=StateBlock)
class StateTestBlockData(StateBlockData):
    def build(self):
        self.x = Var(initialize=0)
        self.pressure = Var()
        self.enth_mol = Var()
        self.temperature = Expression(expr=self.enth_mol * 2.0)
        self.div0 = Expression(expr=4.0 / self.x)
        self.flow_mol = Var(["CO2", "H2O"])


@declare_process_block_class("StateBlockForTesting2", block_class=StateBlock)
class StateTestBlockData(StateBlockData):
    def build(self):
        self.pressure = Var()
        self.temperature = Var()
        self.flow_mol = Var()
        self.flow_vol = Var()


@pytest.fixture
def gtmodel():
    time_set = set([0, 1])
    m = ConcreteModel()
    m.state_a = StateBlockForTesting(time_set)
    m.state_b = StateBlockForTesting(time_set, [1, 2, 3])
    m.state_c = StateBlockForTesting2(time_set)

    m.state_a[0].pressure = 11000
    m.state_a[0].enth_mol = 1100
    m.state_a[0].flow_mol["CO2"] = 110
    m.state_a[0].flow_mol["H2O"] = 111

    m.state_b[0, 1].pressure = 10000
    m.state_b[0, 1].enth_mol = 1000
    m.state_b[0, 1].flow_mol["CO2"] = 100
    m.state_b[0, 1].flow_mol["H2O"] = 101

    m.state_b[0, 2].pressure = 20000
    m.state_b[0, 2].enth_mol = 2000
    m.state_b[0, 2].flow_mol["CO2"] = 200
    m.state_b[0, 2].flow_mol["H2O"] = 201

    m.state_b[0, 3].pressure = 30000
    m.state_b[0, 3].enth_mol = 3000
    m.state_b[0, 3].flow_mol["CO2"] = 300
    m.state_b[0, 3].flow_mol["H2O"] = 301

    m.state_c[0].pressure = 1000
    m.state_c[0].flow_mol = 10
    m.state_c[0].flow_vol = 30
    m.state_c[0].temperature = 300

    return m


@pytest.mark.unit
def test_generate_table(gtmodel):
    m = gtmodel

    sd = {"a": m.state_a[0], "b1": m.state_b[(0, 1)]}
    # This tests what happens if one of the requested attributes gives a division
    # by zero error and if one of the attributes doesn't exist.  With flow it
    # tests indexed attributes.
    st = generate_table(
        sd,
        attributes=[
            "pressure",
            "temperature",
            "div0",
            ("flow_mol", "CO2"),
            ("flow_mol", "H2O"),
            "not_there",
            ("not_there_array", "hi"),
        ],
        heading=["P", "T", "ERR", "F_CO2", "F_H2O", "Miss", "Miss[hi]"],
    )

    assert st.loc["a"]["P"] == 11000
    assert st.loc["a"]["F_CO2"] == 110
    assert st.loc["a"]["F_H2O"] == 111
    assert st.loc["a"]["T"] == 1100 * 2
    assert isna(st.loc["a"]["ERR"])
    assert isna(st.loc["a"]["Miss"])

    assert st.loc["b1"]["P"] == 10000
    assert st.loc["b1"]["F_CO2"] == 100
    assert st.loc["b1"]["F_H2O"] == 101
    assert st.loc["b1"]["T"] == 1000 * 2
    assert isna(st.loc["b1"]["ERR"])
    assert isna(st.loc["b1"]["Miss"])


@pytest.mark.unit
def test_generate_table_errors(gtmodel):
    m = gtmodel

    sd = {"a": m.state_a[0], "b1": m.state_b[0, 1]}
    heading = ["F"]

    with pytest.raises(AssertionError):
        st = generate_table(sd, attributes=[("flow_mol",)], heading=heading)

    with pytest.raises(KeyError):
        st = generate_table(sd, attributes=[("flow_mol", "coffee")], heading=heading)

    with pytest.raises(TypeError):
        st = generate_table(sd, attributes=["flow_mol"], heading=heading)


@pytest.mark.unit
def test_mixed_table(gtmodel):
    m = gtmodel

    sd = {
        "a": m.state_a[0],
        "b[1]": m.state_b[(0, 1)],
        "b[2]": m.state_b[(0, 2)],
        "b[3]": m.state_b[(0, 3)],
        "c": m.state_c[0],
    }

    st = generate_table(
        sd,
        exception=False,
        attributes=(
            "pressure",
            "temperature",
            "flow_mol",
            "flow_vol",
            "enth_mol",
            ("flow_mol", "CO2"),
            ("flow_mol", "H2O"),
        ),
        heading=(
            "P",
            "T",
            "F",
            "Fvol",
            "h",
            "F[CO2]",
            "F[H2O]",
        ),
    )

    assert st.loc["a"]["P"] == 11000
    assert st.loc["a"]["F[CO2]"] == 110
    assert st.loc["a"]["F[H2O]"] == 111
    assert isna(st.loc["a"]["Fvol"])
    assert isna(st.loc["a"]["F"])
    assert st.loc["a"]["T"] == 1100 * 2

    assert st.loc["c"]["P"] == 1000
    assert isna(st.loc["c"]["F[CO2]"])
    assert isna(st.loc["c"]["F[H2O]"])
    assert st.loc["c"]["Fvol"] == 30
    assert st.loc["c"]["F"] == 10
    assert st.loc["c"]["T"] == 300


@pytest.fixture()
def HX1D_array_model():
    # An example of maximum perversity. Dynamics, 1D control volumes, and
    # an indexed unit
    unit_set = range(3)
    time_set = [0, 5]
    time_units = pyunits.s
    fs_cfg = {"dynamic": True, "time_set": time_set, "time_units": time_units}

    m = ConcreteModel()
    m.fs = FlowsheetBlock(**fs_cfg)

    m.fs.properties = thermo_props.SaponificationParameterBlock()

    m.fs.unit_array = HX1D(
        unit_set,
        hot_side_name="shell_side",
        cold_side_name="tube_side",
        shell_side={"property_package": m.fs.properties},
        tube_side={"property_package": m.fs.properties},
    )

    def tube_stream_array_rule(b, i):
        return {
            "source": b.unit_array[i].tube_side_outlet,
            "destination": b.unit_array[i + 1].tube_side_inlet,
        }

    m.fs.tube_stream_array = Arc(range(2), rule=tube_stream_array_rule)

    def shell_stream_array_rule(b, i):
        return {
            "source": b.unit_array[i + 1].shell_side_outlet,
            "destination": b.unit_array[i].shell_side_inlet,
        }

    m.fs.shell_stream_array = Arc(range(1, -1, -1), rule=shell_stream_array_rule)

    TransformationFactory("network.expand_arcs").apply_to(m)
    return m


@pytest.mark.unit
def test_create_stream_table_dataframe_from_Port_HX1D(HX1D_array_model):
    m = HX1D_array_model
    df = create_stream_table_dataframe({"state": m.fs.unit_array[0].tube_side_inlet})

    assert df.loc["Pressure"]["state"] == pytest.approx(101325)
    assert df.loc["Temperature"]["state"] == pytest.approx(298.15)
    assert df.loc["Volumetric Flowrate"]["state"] == pytest.approx(1.0)
    assert df.loc["Molar Concentration H2O"]["state"] == pytest.approx(100.0)
    assert df.loc["Molar Concentration NaOH"]["state"] == pytest.approx(100.0)
    assert df.loc["Molar Concentration EthylAcetate"]["state"] == pytest.approx(100.0)
    assert df.loc["Molar Concentration SodiumAcetate"]["state"] == pytest.approx(100.0)
    assert df.loc["Molar Concentration Ethanol"]["state"] == pytest.approx(100.0)

    df = create_stream_table_dataframe({"state": m.fs.unit_array[0].tube_side_outlet})

    assert df.loc["Pressure"]["state"] == pytest.approx(101325)
    assert df.loc["Temperature"]["state"] == pytest.approx(298.15)
    assert df.loc["Volumetric Flowrate"]["state"] == pytest.approx(1.0)
    assert df.loc["Molar Concentration H2O"]["state"] == pytest.approx(100.0)
    assert df.loc["Molar Concentration NaOH"]["state"] == pytest.approx(100.0)
    assert df.loc["Molar Concentration EthylAcetate"]["state"] == pytest.approx(100.0)
    assert df.loc["Molar Concentration SodiumAcetate"]["state"] == pytest.approx(100.0)
    assert df.loc["Molar Concentration Ethanol"]["state"] == pytest.approx(100.0)


@pytest.mark.unit
def test_create_stream_table_dataframe_from_Arc_HX1D(HX1D_array_model):
    m = HX1D_array_model
    df = create_stream_table_dataframe({"state": m.fs.tube_stream_array}, time_point=5)

    for i in range(2):
        stg = f"state[{i}]"
    assert df.loc["Pressure"][stg] == pytest.approx(101325)
    assert df.loc["Temperature"][stg] == pytest.approx(298.15)
    assert df.loc["Volumetric Flowrate"][stg] == pytest.approx(1.0)
    assert df.loc["Molar Concentration H2O"][stg] == pytest.approx(100.0)
    assert df.loc["Molar Concentration NaOH"][stg] == pytest.approx(100.0)
    assert df.loc["Molar Concentration EthylAcetate"][stg] == pytest.approx(100.0)
    assert df.loc["Molar Concentration SodiumAcetate"][stg] == pytest.approx(100.0)
    assert df.loc["Molar Concentration Ethanol"][stg] == pytest.approx(100.0)


@pytest.fixture()
def flash_model():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.properties = PhysicalParameterTestBlock()  # default={"valid_phase":
    # ('Liq', 'Vap')})
    m.fs.flash = Flash(property_package=m.fs.properties)

    return m


@pytest.mark.unit
def test_state_block_retrieval_fail(flash_model):
    # The flash unit does not have real state blocks associated with the
    # outlet ports. There is a mixture of references, expressions, and multiple
    # state blocks. Therefore we don't want any state block getting through.
    m = flash_model
    with pytest.raises(
        RuntimeError,
        match="No block could be retrieved from Port fs.flash.liq_outlet "
        "because components are derived from multiple blocks.",
    ):
        df = create_stream_table_dataframe({"state": m.fs.flash.liq_outlet})


@pytest.mark.unit
def test_state_block_retrieval_empty_port():
    m = ConcreteModel()
    m.p = Port()
    with pytest.raises(
        ValueError,
        match="No block could be retrieved from Port p because it contains "
        "no components.",
    ):
        df = create_stream_table_dataframe({"state": m.p})


@pytest.fixture()
def distillation_model():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.properties = BTXParameterBlock(
        valid_phase=("Liq", "Vap"), activity_coeff_model="Ideal", state_vars="FTPz"
    )
    m.fs.unit = TrayColumn(
        number_of_trays=3,
        feed_tray_location=2,
        condenser_type=CondenserType.totalCondenser,
        condenser_temperature_spec=TemperatureSpec.atBubblePoint,
        property_package=m.fs.properties,
    )
    return m


@pytest.mark.unit
def test_extended_port_retrieval(distillation_model):
    m = distillation_model
    df = create_stream_table_dataframe({"state": m.fs.unit.feed})
    stg = "state"
    assert df.loc["pressure"][stg] == pytest.approx(101325)
    assert df.loc["temperature"][stg] == pytest.approx(298.15)
    assert df.loc["mole_frac_comp benzene"][stg] == pytest.approx(0.5)
    assert df.loc["mole_frac_comp toluene"][stg] == pytest.approx(0.5)
    assert df.loc["flow_mol"][stg] == pytest.approx(1)
